from typing import Dict, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from ..utils import NORM_LOOKUP, ndarray_to_tensor

__all__ = ["Predictor"]


def compute_pyramid_patch_weight_loss(width: int, height: int) -> np.ndarray:
    """Compute a weight matrix.

    The matrix assigns bigger weight on pixels in the center and less weight to pixels
    on the image boundary. Helps dealing with prediction artifacts on tile boundaries.

    Ported from: pytorch-toolbelt

    Parameters
    ----------
        width : int
            Tile width.
        height : int
            Tile height

    Returns
    -------
        np.ndarray:
            Weight matrix. Shape (H, W).
    """
    xc = width * 0.5
    yc = height * 0.5
    xl = 0
    xr = width
    yb = 0
    yt = height
    Dc = np.zeros((width, height))
    De = np.zeros((width, height))

    Dcx = np.square(np.arange(width) - xc + 0.5)
    Dcy = np.square(np.arange(height) - yc + 0.5)
    Dc = np.sqrt(Dcx[np.newaxis].transpose() + Dcy)

    De_l = np.square(np.arange(width) - xl + 0.5) + np.square(0.5)
    De_r = np.square(np.arange(width) - xr + 0.5) + np.square(0.5)
    De_b = np.square(0.5) + np.square(np.arange(height) - yb + 0.5)
    De_t = np.square(0.5) + np.square(np.arange(height) - yt + 0.5)

    De_x = np.sqrt(np.minimum(De_l, De_r))
    De_y = np.sqrt(np.minimum(De_b, De_t))
    De = np.minimum(De_x[np.newaxis].transpose(), De_y)

    alpha = (width * height) / np.sum(np.divide(De, np.add(Dc, De)))
    W = alpha * np.divide(De, np.add(Dc, De))
    return W


class Predictor:
    def __init__(
        self,
        model: nn.Module,
        patch_size: Tuple[int, int],
        normalization: str = None,
        mean: np.ndarray = None,
        std: np.ndarray = None,
        device: Union[str, torch.device] = "cuda",
    ) -> None:
        """Predict dense soft masks with this helper class.

        Includes a weight matrix that can assign bigger weight on pixels
        in center and less weight to pixels on image boundary. helps
        dealing with prediction artifacts on tile boundaries.

        Parameters
        ----------
            model : nn.Module:
                nn.Module pytorch model
            patch_size : Tuple[int, int]
                The height and width of the input patch.
            normalization : str, optional
                Apply img normalization (Same as during training). One of "dataset",
                "minmax", "norm", "percentile", None.
            mean : Tuple[float, float, float]
                Means for each channel. Used only if `normalization` == "dataset"
            std : Tuple[float, float, float]
                Stds for each channel. Used only if `normalization` == "dataset"
            device : str or torch.device, default="cuda"
                The device of the model. One of "cpu", "cuda".
        """
        allowed = list(NORM_LOOKUP.keys()) + [None]
        if normalization not in allowed:
            raise ValueError(
                f"Illegal normalization. Got: {normalization}. Allowed: {allowed}"
            )

        self.normalization = normalization
        self.mean = mean
        self.std = std

        self.model = model
        self.model.eval()

        weight_mat = compute_pyramid_patch_weight_loss(patch_size[0], patch_size[1])
        self.weight_mat = (
            torch.from_numpy(weight_mat).float().to(device).unsqueeze(0).unsqueeze(0)
        )

    def forward_pass(
        self, patch: Union[np.ndarray, torch.Tensor], in_dim_format: str = "HWC"
    ) -> Dict[str, torch.Tensor]:
        """Input an image patch or batch of patches to the network and return logits.

        Parameters
        ----------
            patch : np.ndarray or torch.Tensor
                Image patch.
            in_dim_format : str, default="HWC"
                The order of the dimensions in the input array.
                One of: "HWC", "BHWC"


        Returns
        -------
            Dict[str, torch.Tensor]:
                A dictionary of output name mapped to output tensors.
        """
        dim_types = ("HWC", "BHWC")
        if in_dim_format not in dim_types:
            raise ValueError(
                f"Illegal `in_dim_format`. Got {in_dim_format}. Allowed: {dim_types}"
            )

        if isinstance(patch, np.ndarray):
            patch = ndarray_to_tensor(patch, in_dim_format, "BCHW")

        if self.normalization is not None:
            kwargs = {}
            if self.mean is not None or self.std is not None:
                kwargs = {"mean": self.mean, "std": self.std}

            patch = NORM_LOOKUP[self.normalization](patch, **kwargs)
        else:
            patch = patch.float()

        with torch.no_grad():
            out = self.model(patch)

        return out

    def classify(
        self,
        patch: torch.Tensor,
        act: Union[str, None] = "softmax",
        apply_weights: bool = False,
    ) -> torch.Tensor:
        """Take in logits and output probabilities.

        Additionally apply a weight matrix to help with boundary artefacts.

        Parameters
        ----------
            patch : torch.Tensor
                A tensor of logits produced by the network. Shape: (B, C, H, W)
            act : str, default="softmax"
                Activation to be used. One of: "sigmoid", "softmax" or None
            apply_weights : bool, default=False
                Apply a weight matrix that assigns bigger weight on pixels in center and
                less weight to pixels on the image boundary.

        Returns
        -------
            torch.Tensor:
                The model prediction. Same shape as input `patch`.
        """
        allowed = ("sigmoid", "softmax", "tanh", None)
        if act not in allowed:
            raise ValueError(
                f"Illegal activation func given. Got: {act}. Allowed: {allowed}"
            )

        # Add weights to pred matrix
        if apply_weights:
            # work out the tensor shape first for the weight mat
            B, C = patch.shape[:2]
            W = torch.repeat_interleave(
                self.weight_mat,
                dim=1,
                repeats=C,
            ).repeat_interleave(repeats=B, dim=0)
            patch *= W

        # apply classification activation
        if act == "sigmoid":
            pred = torch.sigmoid(patch)
        elif act == "tanh":
            pred = torch.tanh(patch)
        elif act == "softmax":
            pred = F.softmax(patch, dim=1)
        else:
            pred = patch

        return pred



from pathlib import Path
from typing import Union

import torch
from torch.utils.data import Dataset

from ..utils import FileHandler

SUFFIXES = (".jpeg", ".jpg", ".tif", ".tiff", ".png")


__all__ = ["FolderDataset"]


class FolderDataset(Dataset, FileHandler):
    def __init__(
        self, path: Union[str, Path], pattern: str = "*", n_images: int = None
    ) -> None:
        """Folder dataset that can be used during inference for loading images.

        NOTE: loads only images.

        Parameters
        ----------
            path : str | Path
                Path to the folder containing image files.
            pattern: str, default="*"
                File pattern for filtering only the files that contain the pattern.
            n_images : int, optional
                First n-number of images used from the folder.

        Raises
        ------
            ValueError if `path` does not exist.
            ValueError if `path` is not a folder.
            ValueError if `path` contains images with illegal suffices.
        """
        super().__init__()

        folder_path = Path(path)
        if not folder_path.exists():
            raise ValueError(f"folder: {folder_path} does not exist")

        if not folder_path.is_dir():
            raise ValueError(f"path: {folder_path} is not a folder")

        if not all([f.suffix in SUFFIXES for f in folder_path.iterdir()]):
            raise ValueError(f"files formats in given folder need to be in {SUFFIXES}")

        self.fnames = sorted(folder_path.glob(pattern))
        if n_images is not None:
            self.fnames = self.fnames[:n_images]

    def __len__(self) -> int:
        """Length of folder."""
        return len(self.fnames)

    def __getitem__(self, index: int) -> torch.Tensor:
        """Read image."""
        fn = self.fnames[index]
        im = FileHandler.read_img(fn.as_posix())
        im = torch.from_numpy(im.transpose(2, 0, 1))

        return {"im": im, "file": fn.name[:-4]}



from .attention_modules import Attention
from .base_modules import Activation, Conv, Identity, Norm, Up
from .conv_block import ConvBlock
from .conv_layer import ConvLayer

__all__ = [
    "Identity",
    "Conv",
    "Up",
    "Activation",
    "Norm",
    "Attention",
    "ConvBlock",
    "ConvLayer",
]



from copy import deepcopy

import pytest
import pytorch_lightning as pl

from cellseg_models_pytorch.datamodules.custom_datamodule import CustomDataModule
from cellseg_models_pytorch.datasets import SegmentationFolderDataset
from cellseg_models_pytorch.models import cellpose_plus
from cellseg_models_pytorch.training.lit import SegmentationExperiment


# @pytest.mark.parametrize
def test_training(img_patch_dir, mask_patch_dir):
    train_ds = SegmentationFolderDataset(
        path=img_patch_dir.as_posix(),
        mask_path=mask_patch_dir.as_posix(),
        img_transforms=["blur"],
        inst_transforms=["cellpose"],
        return_sem=True,
        return_type=True,
        return_inst=False,
        return_weight=False,
        normalization="percentile",
    )
    valid_ds = deepcopy(train_ds)
    test_ds = deepcopy(train_ds)

    datamodule = CustomDataModule(
        [train_ds, valid_ds, test_ds], batch_size=1, num_workers=1
    )

    model = cellpose_plus(sem_classes=7, type_classes=7)
    experiment = SegmentationExperiment(
        model=model,
        branch_losses={"cellpose": "mse_ssim", "sem": "ce_dice", "type": "ce_dice"},
        branch_metrics={"cellpose": [None], "sem": ["miou"], "type": ["miou"]},
        lookahead=False,
    )

    trainer = pl.Trainer(
        max_epochs=1,
        gpus=0,
        profiler="simple",
        move_metrics_to_cpu=True,
        fast_dev_run=True,
    )

    trainer.fit(model=experiment, datamodule=datamodule)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from pathos.multiprocessing import ThreadPool as Pool

from ...transforms.albu_transforms import IMG_TRANSFORMS, compose
from ...utils import FileHandler, TilerStitcher, fix_duplicates

__all__ = ["BaseWriter"]

IMG_SUFFIXES = (".jpeg", ".jpg", ".tif", ".tiff", ".png")
MASK_SUFFIXES = (".mat",)


class BaseWriter(ABC):
    def __init__(
        self,
        in_dir_im: str,
        in_dir_mask: str,
        patch_size: Tuple[int, int],
        stride: int,
        transforms: Optional[List[str]] = None,
        **kwargs,
    ) -> None:
        """Init base class for sliding window data writers."""
        self.im_dir = Path(in_dir_im)
        self.mask_dir = Path(in_dir_mask)
        self.stride = stride
        self.patch_size = patch_size

        if not self.im_dir.exists():
            raise ValueError(f"folder: {self.im_dir} does not exist")

        if not self.im_dir.is_dir():
            raise ValueError(f"path: {self.im_dir} is not a folder")

        if not all([f.suffix in IMG_SUFFIXES for f in self.im_dir.iterdir()]):
            raise ValueError(
                f"files formats in given folder need to be in {IMG_SUFFIXES}"
            )

        if not self.mask_dir.exists():
            raise ValueError(f"folder: {self.mask_dir} does not exist")

        if not self.mask_dir.is_dir():
            raise ValueError(f"path: {self.mask_dir} is not a folder")

        if not all([f.suffix in MASK_SUFFIXES for f in self.mask_dir.iterdir()]):
            raise ValueError(
                f"files formats in given folder need to be in {MASK_SUFFIXES}"
            )

        self.fnames_im = sorted(self.im_dir.glob("*"))
        self.fnames_mask = sorted(self.mask_dir.glob("*"))
        if len(self.fnames_im) != len(self.fnames_mask):
            raise ValueError(
                f"Found different number of files in {self.im_dir.as_posix()} and "
                f"{self.mask_dir.as_posix()}."
            )

        self.transforms = None
        if transforms is not None:
            allowed = list(IMG_TRANSFORMS.keys())
            if not all([tr in allowed for tr in transforms]):
                raise ValueError(
                    f"Wrong transformation. Got: {transforms}. Allowed: {allowed}."
                )

            self.transforms = compose(
                [IMG_TRANSFORMS[tr](**kwargs) for tr in transforms]
            )

    @abstractmethod
    def write(self):
        """Patch images and mask to and write them to disk."""
        raise NotImplementedError

    def _get_tiles(
        self,
        img_path: Union[str, Path],
        mask_path: Union[str, Path],
        pre_proc: Callable = None,
    ) -> Dict[str, np.ndarray]:
        """Read one image and corresponding masks and do tiling on them."""
        # im, masks = self._get_arrays()
        im = FileHandler.read_img(img_path)
        masks = FileHandler.read_mask(mask_path, return_all=True)

        if pre_proc is not None:
            masks = pre_proc(masks)

        inst = None
        types = None
        sem = None
        if "inst_map" in masks.keys():
            inst = masks["inst_map"]
        if "type_map" in masks.keys():
            types = masks["type_map"]
        if "sem_map" in masks.keys():
            sem = masks["sem_map"]

        im_tiler = TilerStitcher(
            im_shape=im.shape, patch_shape=self.patch_size + (3,), stride=self.stride
        )

        mask_tiler = TilerStitcher(
            im_shape=inst.shape, patch_shape=self.patch_size + (1,), stride=self.stride
        )

        tiles = {}
        tiles["image"] = im_tiler.patch(im)
        if inst is not None:
            tiles["inst_map"] = self._fix_duplicates(mask_tiler.patch(inst).squeeze())
        if types is not None:
            tiles["type_map"] = mask_tiler.patch(types).squeeze()
        if sem is not None:
            tiles["sem_map"] = mask_tiler.patch(sem).squeeze()

        if self.transforms is not None:
            tiles = self._transform(tiles)

        return tiles

    def _transform(self, tiles: Dict[str, np.ndarray]) -> np.ndarray:
        """Apply transformations to the tiles one by one."""
        n_tiles = tiles["image"].shape[0]
        masks = [arr for key, arr in tiles.items() if key != "image"]
        mask_names = [key for key in tiles.keys() if key != "image"]

        out_tiles = {k: [] for k in tiles.keys()}
        for i in range(n_tiles):
            m = [mask[i] for mask in masks]
            out = self.transforms(image=tiles["image"][i], masks=m)
            out_tiles["image"].append(out["image"])

            for j, mname in enumerate(mask_names):
                out_tiles[mname].append(out["masks"][j])

        for k, mask in out_tiles.items():
            out_tiles[k] = np.array(mask)

        return out_tiles

    def _fix_duplicates(self, patches_inst: np.ndarray) -> np.ndarray:
        """Fix repeatded labels in a patched instance labelled mask."""
        insts = []

        for i in range(patches_inst.shape[0]):
            insts.append(fix_duplicates(patches_inst[i]))

        insts = np.array(insts)

        return insts.astype("int32")

    def _write_parallel(self, write_func: Callable, arg_list: List) -> None:
        """Write patches in parallel."""
        with Pool() as pool:
            it = pool.imap(write_func, arg_list)
            for _ in it:
                pass

    def _write_sequential(self, write_func: Callable, arg_list: List) -> None:
        """Write patches sequentially."""
        for kwargs in arg_list:
            write_func(**kwargs)



from .custom_datamodule import CustomDataModule
from .downloader import SimpleDownloader
from .lizard_datamodule import LizardDataModule
from .pannuke_datamodule import PannukeDataModule

__all__ = [
    "CustomDataModule",
    "SimpleDownloader",
    "PannukeDataModule",
    "LizardDataModule",
]



import numpy as np
import pytest
import torch

from cellseg_models_pytorch.utils import (
    TilerStitcher,
    TilerStitcherTorch,
    extract_patches_numpy,
    extract_patches_torch,
    stitch_patches_numpy,
    stitch_patches_torch,
)


@pytest.fixture
def rand_tensor():
    return torch.rand([2, 3, 410, 520])


@pytest.mark.parametrize(
    "patch_shape,multichannel",
    [
        ((256, 256, 3), True),
        ((320, 320, 1), False),
        pytest.param((256, 256), False, marks=pytest.mark.xfail),
        pytest.param((256, 256, 3), False, marks=pytest.mark.xfail),
    ],
)
@pytest.mark.parametrize("padding", [True, False])
def test_extract_patches_numpy(img_sample, patch_shape, multichannel, padding):
    im = img_sample

    if not multichannel:
        im = im[..., 0]

    patches, _, _ = extract_patches_numpy(
        im, stride=80, patch_shape=patch_shape, padding=padding
    )

    assert patches.shape[1:] == patch_shape


def test_stitch_patches_numpy(img_sample):
    expected = img_sample

    patch_shape = (128, 128, 3)
    patches, ny, nx = extract_patches_numpy(
        expected, stride=80, patch_shape=patch_shape, padding=True
    )

    observed = stitch_patches_numpy(patches, expected.shape, ny, nx, 80, padding=True)

    assert expected.shape == observed.shape
    np.testing.assert_array_equal(expected, observed)


@pytest.mark.parametrize(
    "patch_shape,multichannel",
    [
        ((256, 256, 3), True),
        ((320, 320, 1), False),
    ],
)
def test_tilerstitcher(img_sample, patch_shape, multichannel):
    expected = img_sample

    if not multichannel:
        expected = expected[..., 0]

    tiler = TilerStitcher(
        im_shape=expected.shape, patch_shape=patch_shape, stride=256, padding=True
    )

    patches = tiler.patch(expected)
    observed = tiler.backstitch(patches)

    assert expected.shape == observed.shape
    np.testing.assert_array_equal(expected, observed)


@pytest.mark.parametrize("padding", [True, False])
def test_extract_patches_torch(rand_tensor, padding):
    out = extract_patches_torch(rand_tensor, 256, (256, 256), padding)

    assert out.shape[-2:] == torch.Size([256, 256])


@pytest.mark.parametrize("padding", [True, False])
def test_stitch_patches_torch(rand_tensor, padding):
    stride = 256
    patches = extract_patches_torch(rand_tensor, stride, (256, 256), padding)
    stitched = stitch_patches_torch(patches, tuple(rand_tensor.shape), stride, padding)

    assert stitched.shape == rand_tensor.shape


@pytest.mark.parametrize("padding", [True, False])
def test_tilerstitchertorch(rand_tensor, padding):
    ts = TilerStitcherTorch(tuple(rand_tensor.shape), (256, 256), 256, padding)

    patches = ts.patch(rand_tensor)
    stitched = ts.backstitch(patches)

    assert stitched.shape == rand_tensor.shape



from typing import Callable, Dict, List

import numpy as np
from pathos.multiprocessing import ThreadPool as Pool
from skimage.util import img_as_ubyte
from tqdm import tqdm

from ..postproc import POSTPROC_LOOKUP
from ..utils import (
    fill_holes_semantic,
    majority_vote_parallel,
    majority_vote_sequential,
    med_filt_parallel,
    med_filt_sequential,
    remove_debris_semantic,
)

__all__ = ["PostProcessor"]


INST_ARGMAX = {
    "stardist": False,
    "stardist_orig": False,
    "cellpose": True,
    "cellpose_old": True,
    "omnipose": True,
    "hovernet": True,
    "drfns": True,
    "dcan": False,
    "dran": False,
}


class PostProcessor:
    def __init__(
        self,
        instance_postproc: str,
        inst_key: str,
        aux_key: str,
        type_post_proc: Callable = None,
        sem_post_proc: Callable = None,
        **kwargs,
    ) -> None:
        """Multi-threaded post-processing.

        Parameters
        ----------
            instance_postproc : str
                The post-processing method for the instance segmentation mask.
            inst_key : str
                The key/name of the model instance prediction output that is used
                for the instance segmentation post-processing pipeline.
            aux_key : Tuple[str, ...]:
                The key/name of the model auxilliary output that is used for the
                instance segmentation post-processing pipeline.
            type_post_proc : Callable, optional
                A post-processing function for the type maps. If not None, overrides
                the default.
            sem_post_proc : Callable, optional
                A post-processing function for the semantc seg maps. If not None,
                overrides the default.
            **kwargs
                Arbitrary keyword arguments that can be used for any of the private
                post-processing functions of this class.
        """
        allowed = list(POSTPROC_LOOKUP.keys())
        if instance_postproc not in allowed:
            raise ValueError(
                f"Illegal post proc method. Got {instance_postproc}. Allowed: {allowed}"
            )

        self.postproc_method = POSTPROC_LOOKUP[instance_postproc]
        self.use_argmax = INST_ARGMAX[instance_postproc]
        self.inst_key = inst_key
        self.aux_key = aux_key
        self.kwargs = kwargs
        self.sem_post_proc = sem_post_proc
        self.type_post_proc = type_post_proc

    def _get_sem_map(
        self,
        prob_map: np.ndarray,
        parallel: bool = False,
        kernel_width: int = 15,
        **kwargs,
    ) -> np.ndarray:
        """Run the semantic segmentation post-processing."""
        sem_map = img_as_ubyte(prob_map)

        if self.sem_post_proc is not None:
            sem = self.sem_post_proc(sem_map)
        else:
            if parallel:
                sem = med_filt_parallel(
                    sem_map, kernel_size=(kernel_width, kernel_width)
                )
            else:
                sem = med_filt_sequential(sem_map, kernel_width)

            sem = np.argmax(sem, axis=0)
            sem = remove_debris_semantic(sem)
            sem = fill_holes_semantic(sem)

        return sem

    def _get_inst_map(
        self, prob_map: np.ndarray, aux_map: np.ndarray, **kwargs
    ) -> np.ndarray:
        """Run the instance labelling post processing."""
        if self.use_argmax:
            inst = np.argmax(prob_map, axis=0)
        else:
            inst = prob_map.squeeze()

        return self.postproc_method(inst, aux_map, **self.kwargs)

    def _get_type_map(
        self,
        prob_map: np.ndarray,
        inst_map: np.ndarray,
        parallel: bool = True,
        **kwargs,
    ) -> np.ndarray:
        """Run the type map post-processing. Majority voting for each instance."""
        type_map = np.argmax(prob_map, axis=0)

        if self.type_post_proc is not None:
            type_map = self.type_post_proc(type_map, inst_map, **kwargs)
        else:
            if parallel:
                type_map = majority_vote_parallel(type_map, inst_map)
            else:
                type_map = majority_vote_sequential(type_map, inst_map)

        return type_map

    def post_proc_pipeline(
        self, maps: Dict[str, np.ndarray], **kwargs
    ) -> Dict[str, np.ndarray]:
        """Run post-processing for all the model outputs.

        Parameters
        ----------
            maps : Dict[str, np.ndarray]
                model head names mapped to model outputs.
                E.g. {"sem": np.ndarray, "type": np.ndarray, "cellpose": np.ndarray}

        Returns
        -------
            Dict[str, np.ndarray]:
                final output names mapped to the final post-processed outputs.
                E.g. {"sem": np.ndarray, "type": np.ndarray, "inst": np.ndarray}
        """
        res = {}
        if "sem" in maps.keys():
            res["sem"] = self._get_sem_map(maps["sem"], **self.kwargs)

        res["inst"] = self._get_inst_map(
            maps[self.inst_key], maps[self.aux_key], **self.kwargs
        )

        if "type" in maps.keys():
            res["type"] = self._get_type_map(maps["type"], res["inst"], **self.kwargs)
            res["inst"] *= res["type"] > 0

        return res

    def run_parallel(
        self, maps: List[Dict[str, np.ndarray]], progress_bar: bool = False
    ) -> List[Dict[str, np.ndarray]]:
        """Run the full post-processing pipeline in parallel for many model outputs.

        Parameters
        ----------
            maps : List[Dict[str, np.ndarray]]
                The model output map dictionaries in a list.
            progress_bar : bool, default=False
                If True, a tqdm progress bar is shown.

        Returns
        -------
            List[Dict[str, np.ndarray]]:
                The post-processed output map dictionaries in a list.
        """
        seg_results = []
        with Pool() as pool:
            if progress_bar:
                it = tqdm(pool.imap(self.post_proc_pipeline, maps), total=len(maps))
            else:
                it = pool.imap(self.post_proc_pipeline, maps)

            for x in it:
                seg_results.append(x)

        return seg_results



import pytest
import torch

from cellseg_models_pytorch.modules.attention_modules import Attention


@pytest.mark.parametrize("attention", ["se", "scse", "eca", "gc"])
def test_attention_forward(attention):
    att = Attention(attention, in_channels=8)
    input = torch.rand([1, 8, 16, 16])
    output = att(input)

    assert output.shape == input.shape
    assert output.dtype == input.dtype


@pytest.mark.parametrize("attention", ["se", "scse", "eca", "gc", None])
def test_attention_backward(attention):
    att = Attention(attention, in_channels=8)
    input = torch.rand([1, 8, 16, 16])
    output = att(input)

    assert output.shape == input.shape
    assert output.dtype == input.dtype



from typing import Tuple

import torch
import torch.nn as nn

from cellseg_models_pytorch.modules import Conv, Norm

__all__ = ["SumBlock"]


class SumBlock(nn.ModuleDict):
    def __init__(
        self,
        in_channels: int,
        skip_channels: Tuple[int, ...],
        convolution: str = "conv",
        normalization: str = "bn",
        **kwargs,
    ) -> None:
        """Merge by summation.

        Handles clashing channel numbers with a regular conv block.

        Parameters
        ----------
            in_channels : int
                Number of input channels
            skip_channels : Tuple[int, ...]:
                Number of skip channels
            convolution : str
                Name of the convolution method in the downsampling blocks.
            normalization : str
                Name of the normalization method in the downsampling blocks.
        """
        super().__init__()
        self.in_channels = in_channels
        self.skip_channels = skip_channels
        self.pool = [chl != self.in_channels for chl in self.skip_channels]

        # add channel pooling modules if necessary
        for i, needs_pooling in enumerate(self.pool):
            if needs_pooling:
                downsample = nn.Sequential(
                    Conv(
                        convolution,
                        in_channels=skip_channels[i],
                        bias=False,
                        out_channels=in_channels,
                        kernel_size=1,
                        padding=0,
                    ),
                    Norm(normalization, num_features=in_channels),
                )
                self.add_module(f"downsample{i + 1}", downsample)

    @property
    def out_channels(self) -> int:
        """Out channels."""
        return self.in_channels

    def forward(
        self,
        x: torch.Tensor,
        skips: Tuple[torch.Tensor, ...],
    ) -> torch.Tensor:
        """Forward of the sum block.

        Parameters
        ----------
            x : torch.Tensor
                Input tensor. Shape: (B, C, H, W)
            skips : Tuple[torch.Tensor, ...]
                All the skip features in a list. Shapes: (B, C, H, W).

        Returns
        -------
            torch.Tensor:
                The summed output tensor. Shape (B, C, H, W).
        """
        if self.values():
            skips = list(skips)
            for i, needs_pooling in enumerate(self.pool):
                if needs_pooling:
                    skips[i] = self[f"downsample{i + 1}"](skips[i])

        x = torch.stack([x, *skips], dim=0).sum(dim=0)

        return x

    def extra_repr(self) -> str:
        """Add extra info to print."""
        s = "in_channels={in_channels}, skip_channels={skip_channels}"
        s = s.format(**self.__dict__) + f", out_channels={self.out_channels}"
        return s



from typing import Any, Dict, Tuple

__all__ = ["_create_hovernet_args"]


def _create_hovernet_args(
    depth: int,
    n_dense_blocks: Tuple[int, int],
    norm: str,
    act: str,
    conv: str,
    att: str,
    preact: bool,
    preattend: bool,
    use_style: bool,
    merge_policy: str,
    skip_params: Dict[str, Any],
) -> Tuple[Dict[str, Any], ...]:
    """Create the correct args to build HoVerNet architecture."""
    skip_params = skip_params if skip_params is not None else {"k": None}

    d12 = tuple(
        {
            "merge_policy": merge_policy,
            "short_skips": ("basic", "dense", "basic"),
            "block_types": (
                ("basic",),
                ("hover_dense",) * ndb,
                ("basic",),
            ),
            "kernel_sizes": ((3,), (3,) * ndb, (1,)),
            "expand_ratios": ((0.5,), (0.0625,) * ndb, (1.0,)),
            "squeeze_ratio": 2.0,
            "groups": ((1,), (4,) * ndb, (1,)),
            "biases": ((False,), (False,) * ndb, (False,)),
            "normalizations": ((None,), (norm,) * ndb, (None,)),
            "activations": ((None,), (act,) * ndb, (None,)),
            "convolutions": ((conv,), (conv,) * ndb, (conv,)),
            "attentions": ((None,), (None,) * ndb, (att,)),
            "preactivates": ((False,), (preact,) * ndb, (False,)),
            "preattends": ((False,), (False,) * ndb, (preattend,)),
            "use_styles": ((False,), (use_style,) * ndb, (False,)),
            "skip_params": {
                "short_skips": ("basic",),
                "block_types": (("basic",),),
                "convolutions": ((conv,),),
                "normalizations": ((norm,),),
                "activations": ((act,),),
                **skip_params,
            },
        }
        for ndb in n_dense_blocks
    )

    d3 = (
        {
            "merge_policy": merge_policy,
            "short_skips": ("basic",),
            "block_types": (("basic",),),
            "kernel_sizes": ((3,),),
            "expand_ratios": ((1.0,),),
            "squeeze_ratio": 1.0,
            "groups": ((1,),),
            "biases": ((False,),),
            "normalizations": ((None,),),
            "activations": ((None,),),
            "convolutions": ((conv,),),
            "attentions": ((att,),),
            "preactivates": ((False,),),
            "preattends": ((preattend,),),
            "use_styles": ((use_style,),),
            "skip_params": {
                "short_skips": ("basic",),
                "block_types": (("basic",),),
                "convolutions": ((conv,),),
                "normalizations": ((norm,),),
                "activations": ((act,),),
                **skip_params,
            },
        },
    )

    d45 = (
        {
            "merge_policy": merge_policy,
            "short_skips": ("basic",),
            "block_types": (("basic",),),
            "kernel_sizes": ((3,),),
            "expand_ratios": ((1.0,),),
            "squeeze_ratio": 1.0,
            "groups": ((1,),),
            "biases": ((False,),),
            "normalizations": ((norm,),),
            "activations": ((act,),),
            "convolutions": ((conv,),),
            "attentions": ((att,),),
            "preactivates": ((preact,),),
            "preattends": ((preattend,),),
            "use_styles": ((use_style,),),
            "skip_params": {
                "short_skips": ("basic",),
                "block_types": (("basic",),),
                "convolutions": ((conv,),),
                "normalizations": ((norm,),),
                "activations": ((act,),),
                **skip_params,
            },
        },
    ) * (depth - 3)

    return d12 + d3 + d45



from typing import Tuple

import torch
import torch.nn as nn

from ...modules import ConvLayer, Up
from .merging import Merge

__all__ = ["UnetppSkip"]


class UnetppSkip(nn.ModuleDict):
    def __init__(
        self,
        stage_ix: int,
        dec_channels: Tuple[int, ...],
        skip_channels: Tuple[int, ...],
        dec_dims: Tuple[int, ...],
        hid_channels: int = 256,
        n_layers: int = 1,
        n_blocks: Tuple[int, ...] = (1,),
        short_skips: Tuple[str, ...] = ("residual",),
        block_types: Tuple[Tuple[str, ...], ...] = (("basic",),),
        activations: Tuple[Tuple[str, ...], ...] = (("relu",),),
        normalizations: Tuple[Tuple[str, ...], ...] = (("bn",),),
        convolutions: Tuple[Tuple[str, ...], ...] = (("conv",),),
        attentions: Tuple[Tuple[str, ...], ...] = ((None,),),
        kernel_sizes: Tuple[int, ...] = ((3,),),
        groups: Tuple[int, ...] = ((1,),),
        biases: Tuple[Tuple[bool, ...], ...] = ((False,),),
        preactivates: Tuple[Tuple[bool, ...], ...] = ((False,),),
        preattends: Tuple[Tuple[bool, ...], ...] = ((False,),),
        use_styles: Tuple[Tuple[bool, ...], ...] = ((False,),),
        expand_ratios: Tuple[float, float] = ((1.0,),),
        merge_policy: str = "cat",
        **kwargs,
    ) -> None:
        """Unet++-like skip connection block with added flexibility.

        UNet++: A Nested U-Net Architecture for Medical Image Segmentation
            - https://arxiv.org/abs/1807.10165

        Parameters
        ----------
            stage_ix : int
                The index number of the current decoder stage.
            dec_channels : Tuple[int, ...]
                The number of output channels in the decoder output stages. First elem
                is the number of channels in the encoder head or bottleneck.
            skip_channels : Tuple[int, ...]
                List of the number of channels in the encoder skip tensors.
            dec_dims : Tuple[int, ...]
                List of the heights/widths of each encoder/decoder feature map
                e.g. [256, 128, 64, 32, 16]. Assumption is that feature maps are square.
            hid_channels : int, default=256
                Number of output channels from the hidden middle blocks of unet++.
            n_layers : int, default=1
                The number of conv layers inside one skip stage.
            n_blocks : Tuple[int, ...], default=(1, )
                Number of conv-blocks used at each layer of the skip connection.
            short_skips : str, default=("residual", )
                The short skip methods used inside the conv layers.
            block_types : Tuple[Tuple[str, ...], ...], default=(("basic",), )
                The type of the conv blocks in the conv blocks inside the layers.
            normalizations : Tuple[Tuple[str, ...], ...], default: (("bn",), )
                Normalization methods used in the conv blocks inside the conv layers.
            convolutions : Tuple[str, ...], default=(("conv",),)
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            activations : Tuple[Tuple[str, ...], ...], default: (("relu",), )
                Activation methods used inside the conv layers.
            kernel_sizes : Tuple[int, ...], default=(3, 3)
                The size of the convolution kernels in each conv block.
            groups : int, default=(1, 1)
                Number of groups for the kernels in each convolution blocks.
            biases : bool, default=(True, True)
                Include bias terms in the convolution blocks.
            attentions : Tuple[Tuple[str, ...], ...], default: ((None,), )
                Attention methods used inside the conv layers.
            preactivates Tuple[Tuple[bool, ...], ...], default: ((False,), )
                Boolean flags for the conv layers to use pre-activation.
            preattends Tuple[Tuple[bool, ...], ...], default: ((False,), )
                Boolean flags for the conv layers to use pre-activation.
            use_styles : Tuple[Tuple[bool, ...], ...], default=((False,), )
                Boolean flags for the conv layers to add style vectors at each block.
            expand_ratios : Tuple[float, float], default=((1.0, ),)
                Expand ratios for the conv blocks.
            upsampling : str, default="fixed-unpool"
                Name of the upsampling method.
            merge_policy : str, default="sum"
                The long skip merge policy. One of: "sum", "cat"
            lite_version : bool, default=False
                If True, the dense decoder-to-decoder skips are not utilized at all.
                Reduces the model params quite a lot and computational cost.
        """
        super().__init__()
        self.stage_ix = stage_ix
        self.skip_channels = skip_channels

        # hot-fix
        kwargs.pop("in_channels") if "in_channels" in kwargs.keys() else kwargs

        illegal_args = [
            (k, a)
            for k, a in locals().items()
            if isinstance(a, tuple)
            and a not in (skip_channels, dec_channels, dec_dims)
            and len(a) != n_layers
        ]

        if illegal_args:
            raise ValueError(
                f"""
                All the tuple-arg lengths need to be equal to `n_layers`={n_layers}.
                Illegal args: {illegal_args}"""
            )

        if stage_ix < len(skip_channels):

            curr_channels = skip_channels[stage_ix]
            prev_channels = skip_channels[stage_ix - 1]

            mid_channels = []
            fin_channels = [curr_channels]

            for i in range(stage_ix):
                up_scale = Up("fixed-unpool", scale_factor=2)
                self.add_module(f"up_scale{i + 1}", up_scale)

                channels = [prev_channels] if i == 0 else mid_channels + [hid_channels]
                merge = Merge(
                    merge_policy,
                    in_channels=curr_channels,
                    skip_channels=tuple(channels),
                    **kwargs,
                )
                self.add_module(f"mid_merge{i + 1}", merge)

                for j in range(n_layers):
                    layer = ConvLayer(
                        in_channels=merge.out_channels,
                        out_channels=hid_channels,
                        n_blocks=n_blocks[j],
                        expand_ratios=expand_ratios[j],
                        short_skip=short_skips[j],
                        block_types=block_types[j],
                        activations=activations[j],
                        normalizations=normalizations[j],
                        convolutions=convolutions[j],
                        attentions=attentions[j],
                        preattends=preattends[j],
                        preactivates=preactivates[j],
                        use_styles=use_styles[j],
                        kernel_sizes=kernel_sizes[j],
                        groups=groups[j],
                        biases=biases[j],
                    )
                    self.add_module(f"mid_layer{i + 1}", layer)

                # previous mid block channels
                fin_channels.append(layer.out_channels)
                mid_channels.append(layer.out_channels)  # new mid block channels

            final_merge = Merge(
                merge_policy,
                in_channels=dec_channels[stage_ix],
                skip_channels=tuple(fin_channels),
                **kwargs,
            )
            self.add_module("final_merge", final_merge)
        else:
            # place-holder Merge-for the final decoder stage
            self.add_module(
                "final_merge",
                Merge(
                    None,
                    in_channels=dec_channels[stage_ix],
                    out_channels=dec_channels[stage_ix],
                ),
            )

    @property
    def out_channels(self) -> int:
        """Out channels."""
        return self.final_merge.out_channels

    def forward(
        self,
        x: torch.Tensor,
        skips: Tuple[torch.Tensor],
        extra_skips: Tuple[torch.Tensor] = None,
        **kwargs,
    ) -> torch.Tensor:
        """Forward pass of the Unet++ skip connection.

        Parameters
        ----------
            x : torch.Tensor
                Input tensor. Shape (B, C, H, W).
            skips : Tuple[torch.Tensor, ...]
                All of feature maps from consecutive encoder blocks.
                Order is bottom up. Shapes: (B, C, H, W).
            extra_skips : Tuple[torch.Tensor, ...], default=None
                Extra skip connections. (Previous mid features).

        Returns
        -------
            torch.Tensor:
                Output torch.Tensor. Shape: (B, C, H, W).
        """
        mid_features = []
        fin_features = []
        if self.stage_ix < len(skips):
            current_skip = skips[self.stage_ix]
            prev_skip = skips[self.stage_ix - 1]
            fin_features = [current_skip]

            upscales = [k for k in self.keys() if "up_scale" in k]
            merges = [k for k in self.keys() if "mid_merge" in k]
            layers = [k for k in self.keys() if "mid_layer" in k]
            for i, (up, merge, layer) in enumerate(zip(upscales, merges, layers)):
                upskip = prev_skip if i == 0 else extra_skips[i - 1]
                prev_feat = self[up](upskip)

                mid_feat = self[merge](current_skip, [prev_feat] + mid_features)
                mid_feat = self[layer](mid_feat)
                mid_features.append(mid_feat)

            fin_features = fin_features + mid_features

        x = self.final_merge(x, tuple(fin_features))

        return x, mid_features



import pytest
import torch

from cellseg_models_pytorch.datasets.hdf5_dataset import SegmentationHDF5Dataset

img_transforms = ["rigid", "blur"]
inst_transforms = ["smooth_dist"]


# @pytest.mark.optional
@pytest.mark.parametrize("return_inst", [True, False])
@pytest.mark.parametrize("return_type", [True, False])
@pytest.mark.parametrize("return_sem", [True, False])
@pytest.mark.parametrize("normalization", [None, "minmax"])
def test_hdf5_dataset(hdf5db, return_inst, return_type, return_sem, normalization):
    ds = SegmentationHDF5Dataset(
        path=hdf5db,
        img_transforms=img_transforms,
        inst_transforms=inst_transforms,
        normalization=normalization,
        return_inst=return_inst,
        return_type=return_type,
        return_sem=return_sem,
    )

    out = ds.__getitem__(0)

    if return_inst:
        assert "inst" in out.keys()
        assert out["inst"].dtype == torch.int64
    else:
        assert "binary" not in out.keys()

    if return_type:
        assert "type" in out.keys()
        assert out["type"].dtype == torch.int64
    else:
        assert "type" not in out.keys()

    if return_sem:
        assert "sem" in out.keys()
        assert out["sem"].dtype == torch.int64
    else:
        assert "sem" not in out.keys()

    assert "smoothdist" in out.keys()
    assert out["image"].dtype == torch.float32



import torch
import torch.nn as nn

__all__ = ["SegHead"]


class SegHead(nn.Module):
    def __init__(
        self, in_channels: int, out_channels: int, kernel_size: int = 1
    ) -> None:
        """Segmentation head at the end of decoder branches.

        Parameters
        ----------
            in_channels : int
                Number of channels in the input tensor.
            out_channels : int
                Number of channels in the output tensor.
            kernel_size : int, default=1
                Kernel size for the conv operation.

        """
        super().__init__()
        self.n_classes = out_channels

        if kernel_size != 1:
            self.head = nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                padding=kernel_size // 2,
            )
        else:
            self.head = nn.Conv2d(
                in_channels, out_channels, kernel_size=1, padding=0, bias=False
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the segmentation head."""
        return self.head(x)



from typing import Tuple, Union

import torch

__all__ = [
    "dataset_normalize_torch",
    "percentile",
    "percentile_normalize_torch",
    "minmax_normalize_torch",
    "normalize_torch",
    "NORM_LOOKUP",
]


def dataset_normalize_torch(
    img: torch.Tensor,
    mean: Tuple[float, float, float],
    std: Tuple[float, float, float],
    to_uint8: bool = True,
) -> torch.Tensor:
    """Normalize a 3-channel img tensor with mean and standard deviation of the dataset.

    Parameters
    ----------
        img : torch.Tensor
            Tensor img of Shape (C, H, W) or (B, C, H, W).
        mean : Tuple[float, float, float]
            Means for each channel.
        std : Tuple[float, float, float]
            Standard deviations for each channel.
        to_uint8 : bool, default=True
            If input tensor values between [0, 255]. The std and mean
            can be scaled from [0, 1] -> [0, 255].

    Raises
    ------
        TypeError if input img is not torch.Tensor.
        ValueError if input image has illegal shape.
        ZeroDivisionError if normalizing leads to zero-div error.

    Returns
    -------
        torch.Tensor:
            Normalized Tensor image. Same shape as input.
    """
    if not isinstance(img, torch.Tensor):
        raise TypeError(f"input img needs to be a tensor. Got {img.dtype}.")

    if not 3 <= img.ndim <= 4:
        raise ValueError("img tensor shape should be either (C, H, W)|(B, C, H, W)")

    if len(mean) != 3 and len(std) != 3:
        raise ValueError(
            f"Mean and std is needed for every channel. Got: {mean} and {std}"
        )

    img = img.float()
    mean = torch.tensor(mean, dtype=img.dtype, device=img.device)
    std = torch.tensor(std, dtype=img.dtype, device=img.device)

    if to_uint8:
        mean = mean * 255
        std = std * 255

    if (std == 0).any():
        raise ZeroDivisionError(
            "zeros detected in std-values -> would lead to zero-div error"
        )

    if mean.ndim == 1:
        mean = mean.view(-1, 1, 1)
    if std.ndim == 1:
        std = std.view(-1, 1, 1)

    img.sub_(mean).div_(std)

    return img


# Channel-wise normalizations per image.
# Ported from: https://gist.github.com/spezold/42a451682422beb42bc43ad0c0967a30
def percentile(t: torch.tensor, q: float) -> Union[int, float]:
    """Return the ``q``-th percentile of the flattened input tensor's data.

    CAUTION:
     * Needs PyTorch >= 1.1.0, as ``torch.kthvalue()`` is used.
     * Values are not interpolated, which corresponds to
       ``numpy.percentile(..., interpolation="nearest")``.

    :param t: Input tensor.
    :param q: Percentile to compute, which must be between 0 and 100
              inclusive.

    :return: Resulting value (scalar).
    """
    # Note that ``kthvalue()`` works one-based, i.e. the first sorted
    # value indeed corresponds to k=1, not k=0! Use float(q) instead of
    # q directly, so that ``round()`` returns an integer, even if q is
    # a np.float32.
    k = 1 + round(0.01 * float(q) * (t.numel() - 1))
    result = t.contiguous().view(-1).kthvalue(k).values.item()
    return result


def percentile_normalize_torch(img: torch.Tensor) -> torch.Tensor:
    """1-99 percentile normalization per image channel.

    Parameters
    ----------
        img : torch.Tensor
            Input image to be normalized. Shape (C, H, W)|(B, C, H, W).

    Raises
    ------
        ValueError if input image has illegal shape.

    Returns
    -------
        torch.Tensor:
            Normalized image. Shape (C, H, W)|(B, C, H, W).
    """
    if img.ndim == 4:
        _, C, _, _ = img.shape
    elif img.ndim == 3:
        C, _, _ = img.shape
    else:
        raise ValueError("img tensor shape should be either (C, H, W)|(B, C, H, W)")

    img = img.float()
    percentile1 = torch.zeros(C, dtype=img.dtype, device=img.device)
    percentile99 = torch.zeros(C, dtype=img.dtype, device=img.device)
    for channel in range(C):
        if img.ndim == 4:
            percentile1[channel] = percentile(img[:, channel, ...], q=1)
            percentile99[channel] = percentile(img[:, channel, ...], q=99)
        else:
            percentile1[channel] = percentile(img[channel, ...], q=1)
            percentile99[channel] = percentile(img[channel, ...], q=99)

    img.sub_(percentile1.view(-1, 1, 1)).div_(
        (percentile99 - percentile1).view(-1, 1, 1)
    )

    return img


def minmax_normalize_torch(img: torch.Tensor) -> torch.Tensor:
    """Min-max normalize image tensor per channel.

    Parameters
    ----------
        img : torch.Tensor
            Input image tensor. shape (C, H, W)|(B, C, H, W).

    Raises
    ------
        ValueError if input image has illegal shape.

    Returns
    -------
        torch.Tensor:
            Minmax normalized image tensor. Shape (C, H, W)|(B, C, H, W).
    """
    if img.ndim == 4:
        _, C, _, _ = img.shape
    elif img.ndim == 3:
        C, _, _ = img.shape
    else:
        raise ValueError("img tensor shape should be either (C, H, W)|(B, C, H, W)")

    img = img.float()
    chl_min = torch.zeros(C, dtype=img.dtype, device=img.device)
    chl_max = torch.zeros(C, dtype=img.dtype, device=img.device)
    for channel in range(C):
        if img.ndim == 4:
            chl_min[channel] = torch.min(img[:, channel, ...])
            chl_max[channel] = torch.max(img[:, channel, ...])
        else:
            chl_min[channel] = torch.min(img[channel, ...])
            chl_max[channel] = torch.max(img[channel, ...])

    return img.sub_(chl_min.view(-1, 1, 1)).div_((chl_max - chl_min).view(-1, 1, 1))


def normalize_torch(img: torch.Tensor) -> torch.Tensor:
    """Standardize image tensor per channel.

    Parameters
    ----------
        img : torch.Tensor
            Input image tensor. shape (C, H, W).

    Raises
    ------
        ValueError if input image has illegal shape.

    Returns
    -------
        torch.Tensor:
            Standardized image tensor. Shape (C, H, W)|(B, C, H, W).
    """
    if not 3 <= img.ndim <= 4:
        raise ValueError("img tensor shape should be either (C, H, W)|(B, C, H, W)")

    img = img.float()
    chl_means = torch.mean(img, dim=(-2, -1))
    chl_stds = torch.std(img, dim=(-2, -1))

    img.sub_(chl_means.view(-1, 1, 1)).div_(chl_stds.view(-1, 1, 1))

    return img


NORM_LOOKUP = {
    "minmax": minmax_normalize_torch,
    "norm": normalize_torch,
    "percentile": percentile_normalize_torch,
    "dataset": dataset_normalize_torch,
}



from torch.nn import Conv2d

from .scaled_ws_conv import ScaledWSConv2d
from .ws_conv import WSConv2d

CONV_LOOKUP = {"conv": Conv2d, "wsconv": WSConv2d, "scaled_wsconv": ScaledWSConv2d}


__all__ = ["CONV_LOOKUP", "WSConv2d", "ScaledWSConv2d"]



from pathlib import Path
from typing import Callable, Dict, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from ._base_inferer import BaseInferer

__all__ = ["SlidingWindowInferer"]


class SlidingWindowInferer(BaseInferer):
    def __init__(
        self,
        model: nn.Module,
        input_folder: Union[Path, str],
        out_activations: Dict[str, str],
        out_boundary_weights: Dict[str, bool],
        stride: int,
        patch_size: Tuple[int, int],
        instance_postproc: str,
        padding: int = None,
        batch_size: int = 8,
        normalization: str = None,
        device: str = "cuda",
        save_masks: bool = True,
        save_intermediate: bool = False,
        save_dir: Union[Path, str] = None,
        checkpoint_path: Union[Path, str] = None,
        n_images: int = None,
        type_post_proc: Callable = None,
        sem_post_proc: Callable = None,
        **postproc_kwargs,
    ) -> None:
        """Sliding window inference for a folder of images.

        NOTE: This class assumes that all the images in the input folder
        have the same shape. If the images have different shapes,
        `batch_size` needs to be set to 1 for this to work.

        Parameters
        ----------
            model : nn.Module
                A segmentation model.
            input_folder : Path | str
                Path to a folder of images.
            out_activations : Dict[str, str]
                Dictionary of head names mapped to a string value that specifies the
                activation applied at the head. E.g. {"type": "tanh", "cellpose": None}
                Allowed values: "softmax", "sigmoid", "tanh", None.
            out_boundary_weights : Dict[str, bool]
                Dictionary of head names mapped to a boolean value. If the value is
                True, after a prediction, a weight matrix is applied that assigns bigger
                weight on pixels in the center and less weight to pixels on the tile
                boundaries. helps dealing with prediction artefacts on the boundaries.
                E.g. {"type": False, "cellpose": True}
            stride : int
                Stride of the sliding window.
            patch_size : Tuple[int, int]:
                The size of the input patches that are fed to the segmentation model.
            instance_postproc : str
                The post-processing method for the instance segmentation mask. One of:
                "cellpose", "omnipose", "stardist", "hovernet", "dcan", "drfns", "dran"
            padding : int, optional
                The amount of reflection padding for the input images.
            batch_size : int, default=8
                Number of images loaded from the folder at every batch.
            normalization : str, optional
                Apply img normalization (Same as during training). One of "dataset",
                "minmax", "norm", "percentile", None.
            device : str, default="cuda"
                The device of the input and model. One of: "cuda", "cpu"
            save_masks : bool, default=False
                If True, the resulting segmentation masks will be saved into `out_masks`
                variable.
            save_intermediate : bool, default=False
                If True, intermediate soft masks will be saved into `soft_masks` var.
            save_dir : bool, optional
                Path to save directory. If None, no masks will be saved to disk as .mat
                files. If not None, overrides `save_masks`, thus for every batch the
                segmentation results are saved into disk and the intermediate results
                are flushed.
            checkpoint_path : Path | str, optional
                Path to the model weight checkpoints.
            n_images : int, optional
                First n-number of images used from the `ìnput_folder`.
            type_post_proc : Callable, optional
                A post-processing function for the type maps. If not None, overrides
                the default.
            sem_post_proc : Callable, optional
                A post-processing function for the semantc seg maps. If not None,
                overrides the default.
            **postproc_kwargs:
                Arbitrary keyword arguments for the post-processing.
        """
        super().__init__(
            model=model,
            input_folder=input_folder,
            out_activations=out_activations,
            out_boundary_weights=out_boundary_weights,
            patch_size=patch_size,
            padding=padding,
            batch_size=batch_size,
            normalization=normalization,
            instance_postproc=instance_postproc,
            device=device,
            save_masks=save_masks,
            save_intermediate=save_intermediate,
            save_dir=save_dir,
            checkpoint_path=checkpoint_path,
            n_images=n_images,
            type_post_proc=type_post_proc,
            sem_post_proc=sem_post_proc,
            **postproc_kwargs,
        )

        self.stride = stride

    def _get_margins(
        self, first_endpoint: int, img_size: int, stride: int, pad: int = None
    ) -> Tuple[int, int]:
        """Get the number of slices needed for one direction and the overlap."""
        if pad is not None:
            img_size += int(pad)

        n = 1
        mod = 0
        end = first_endpoint
        while True:
            n += 1
            end += stride

            if end > img_size:
                mod = end - img_size
                break
            elif end == img_size:
                break

        return n, mod + pad

    def _get_slices(
        self,
        stride: int,
        patch_size: Tuple[int, int],
        img_size: Tuple[int, int],
        pad: int = None,
    ) -> Tuple[Dict[str, slice], int, int]:
        """Get all the overlapping slices in a dictionary and the needed paddings."""
        y_end, x_end = patch_size
        nrows, pady = self._get_margins(y_end, img_size[0], stride, pad=pad)
        ncols, padx = self._get_margins(x_end, img_size[1], stride, pad=pad)

        xyslices = {}
        for row in range(nrows):
            for col in range(ncols):
                y_start = row * stride
                y_end = y_start + patch_size[0]
                x_start = col * stride
                x_end = x_start + patch_size[1]
                xyslices[f"y-{y_start}_x-{x_start}"] = (
                    slice(y_start, y_end),
                    slice(x_start, x_end),
                )

        return xyslices, pady, padx

    def _infer_batch(self, input_batch: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Infer one batch of images."""
        slices, pady, padx = self._get_slices(
            self.stride, self.patch_size, tuple(input_batch.shape[2:]), self.padding
        )

        padx, modx = divmod(padx, 2)
        pady, mody = divmod(pady, 2)
        padx += modx
        pady += mody

        input_batch = F.pad(
            input_batch.float(), pad=(padx, padx, pady, pady), mode="reflect"
        )

        # initialize the output masks.
        out_maps = {}
        max_channels = 0
        for head_name, out_channels in self.out_heads:
            # use the largest out channel for recovery mask.
            max_channels = out_channels if out_channels > max_channels else max_channels

            out_maps[head_name] = torch.zeros(
                input_batch.shape[0],
                out_channels,
                *input_batch.shape[2:],
                dtype=input_batch.dtype,
                device=self.device,
            )

        # patches are added to the out mask so need a recovery mask.
        recovery = torch.zeros(
            input_batch.shape[0],
            max_channels,
            *input_batch.shape[2:],
            dtype=input_batch.dtype,
            device=self.device,
        )

        # run inference with the slices
        for k, (yslice, xslice) in slices.items():
            batch = input_batch[..., yslice, xslice].to(self.device).float()
            logits = self.predictor.forward_pass(batch)

            probs = {}
            for k, logit in logits.items():
                probs[k] = self.predictor.classify(logit, **self.head_kwargs[k])

            for k, out_map in out_maps.items():
                out_map[..., yslice, xslice] += probs[k]

            recovery[..., yslice, xslice] += 1

        for k, out_map in out_maps.items():
            out = out_map / recovery[:, 0 : out_map.shape[1], ...]
            out_maps[k] = out[..., pady:-pady, padx:-padx]

        return out_maps



from pathlib import Path
from typing import Any, Dict, List, Tuple

import numpy as np
from pathos.multiprocessing import ThreadPool as Pool
from tqdm import tqdm

from ..metrics import (
    accuracy_multiclass,
    aggregated_jaccard_index,
    average_precision,
    dice2,
    dice_multiclass,
    f1score_multiclass,
    iou_multiclass,
    panoptic_quality,
    sensitivity_multiclass,
    specificity_multiclass,
)
from .file_manager import FileHandler
from .mask_utils import get_type_instances, remap_label

__all__ = ["INST_METRIC_LOOKUP", "SEM_METRIC_LOOKUP", "BenchMarker"]


INST_METRIC_LOOKUP = {
    "pq": panoptic_quality,
    "aji": aggregated_jaccard_index,
    "dice2": dice2,
    "ap": average_precision,
}


SEM_METRIC_LOOKUP = {
    "iou": iou_multiclass,
    "dice": dice_multiclass,
    "f1": f1score_multiclass,
    "acc": accuracy_multiclass,
    "sensitivity": sensitivity_multiclass,
    "specificity": specificity_multiclass,
}


class BenchMarker:
    def __init__(
        self, pred_dir: str, true_dir: str, classes: Dict[str, int] = None
    ) -> None:
        """Run benchmarking, given prediction and ground truth mask folders.

        Parameters
        ----------
            pred_dir : str
                Path to the prediction .mat files. The pred files have to have matching
                names to the gt filenames.
            true_dir : str
                Path to the ground truth .mat files. The gt files have to have matching
                names to the pred filenames.
            classes : Dict[str, int], optional
                Class dict. E.g. {"bg": 0, "epithelial": 1, "immmune": 2}
        """
        self.pred_dir = Path(pred_dir)
        self.true_dir = Path(true_dir)
        self.classes = classes

    @staticmethod
    def compute_inst_metrics(
        true: np.ndarray,
        pred: np.ndarray,
        name: str,
        metrics: Tuple[str, ...] = ("pq",),
        type: str = "binary",
    ) -> Dict[str, float]:
        """Compute metrics for one prediciton-ground truth pair.

        Parameters
        ----------
            true : np.ndarray
                The ground truth instance labelled mask. Shape: (H, W).
            pred : np.ndarray
                The predicted instance labelled mask. Shape: (H, W).
            name : str
                Name of the sample/image.
            metrics : Tuple[str, ...], default=("pq", )
                A tuple of the metrics that will be computed. Allowed metrics:
                "pq", "aji", "dice2", "split_and_merge".

        Raises
        ------
            ValueError if an illegal metric is given.

        Returns
        -------
            Dict[str, float]:
                A dictionary where metric names are mapped to metric values.
                e.g. {"pq": 0.5, "aji": 0.55, "name": "sample1"}
        """
        allowed = list(INST_METRIC_LOOKUP.keys())
        if not all([m in allowed for m in metrics]):
            raise ValueError(
                f"An illegal metric was given. Got: {metrics}, allowed: {allowed}"
            )

        # Skip empty GTs
        if len(np.unique(true)) > 1:
            true = remap_label(true)
            pred = remap_label(pred)

            met = {}
            for m in metrics:
                met[m] = INST_METRIC_LOOKUP[m]

            res = {}
            for k, m in met.items():
                score = m(true, pred)

                if isinstance(score, dict):
                    for n, v in score.items():
                        res[n] = v
                else:
                    res[k] = score

            res["name"] = name
            res["type"] = type

            return res

    @staticmethod
    def compute_sem_metrics(
        true: np.ndarray,
        pred: np.ndarray,
        name: str,
        num_classes: int,
        metrics: Tuple[str, ...] = ("iou",),
    ) -> Dict[str, float]:
        """Compute metrics for one prediciton-ground truth pair.

        Parameters
        ----------
            true : np.ndarray
                The ground truth semantic seg mask. Shape: (H, W).
            pred : np.ndarray
                The predicted semantic seg mask. Shape: (H, W).
            name : str
                Name of the sample/image.
            num_classes : int
                Number of classes in the dataset.
            metrics : Tuple[str, ...], default=("iou", )
                A tuple of the metrics that will be computed. Allowed metrics:
                "iou", "aji", "dice2", "split_and_merge".

        Raises
        ------
            ValueError if an illegal metric is given.

        Returns
        -------
            Dict[str, float]:
                A dictionary where metric names are mapped to metric values.
                e.g. {"iou": 0.5, "f1score": 0.55, "name": "sample1"}
        """
        allowed = list(SEM_METRIC_LOOKUP.keys())
        if not all([m in allowed for m in metrics]):
            raise ValueError(
                f"An illegal metric was given. Got: {metrics}, allowed: {allowed}"
            )

        # Skip empty GTs
        met = {}
        for m in metrics:
            met[m] = SEM_METRIC_LOOKUP[m]

        res = {}
        for k, m in met.items():
            score = m(true, pred, num_classes=num_classes)
            res[k] = score

        res["name"] = name
        return res

    def _compute_inst_metrics(
        args: List[Tuple[np.ndarray, np.ndarray, str, List[str]]]
    ) -> Dict[str, float]:
        return BenchMarker.compute_inst_metrics(*args)

    def _compute_sem_metrics(
        args: List[Tuple[np.ndarray, np.ndarray, str, List[str], int]]
    ) -> Dict[str, float]:
        return BenchMarker.compute_sem_metrics(*args)

    @staticmethod
    def run_metrics(
        args: List[Tuple[np.ndarray, np.ndarray, str, List[str]]],
        type: str = "inst",
        msg: str = None,
    ) -> List[Dict[str, float]]:
        """Run segmentation metrics in parallel.

        Parameters
        ----------
            args : List[Tuple[np.ndarray, np.ndarray, str, List[str]]]
                A list of params for `compute_inst_metrics` or `compute_sem_metrics`
            type : str, default="inst"
                One of "inst" or "sem".
            msg : str, optional
                msg for the progress bar.

        Returns
        -------
            List[Dict[str, float]]:
                A list of dictionaries where metric names are mapped to metric values.
        """
        f = (
            BenchMarker._compute_sem_metrics
            if type == "sem"
            else BenchMarker._compute_inst_metrics
        )

        info = "" if msg is None else msg
        metrics = []
        with Pool() as pool:
            for x in tqdm(
                pool.imap_unordered(f, args),
                total=len(args),
                desc=f"Running {info} metrics",
            ):
                metrics.append(x)

        return metrics

    def _read_files(self) -> List[Tuple[np.ndarray, np.ndarray, str]]:
        """Read in the files from the input folders."""
        preds = sorted(self.pred_dir.glob("*"))
        trues = sorted(self.true_dir.glob("*"))

        masks = []
        for truef, predf in zip(trues, preds):
            true = FileHandler.read_mask(truef, return_all=True)
            pred = FileHandler.read_mask(predf, return_all=True)
            name = truef.name
            masks.append((true, pred, name))

        return masks

    def run_inst_benchmark(
        self, how: str = "binary", metrics: Tuple[str, ...] = ("pq",)
    ) -> None:
        """Run instance segmentation benchmarking.

        Parameters
        ----------
            how : str, default="binary"
                One of "binary" or "multi". "multi" computes metrics per class and
                "binary" computes metrics for all objects.
            metrics : Tuple[str, ...], default=("pq", )
                A tuple of the metrics that will be computed. Allowed metrics:
                "pq", "aji", "dice2" "ap".

        Raises
        ------
            ValueError if an illegal `how` arg is given.

        Returns
        -------
            Dict[str, Any]:
                Dictionary mapping the metrics to values + metadata.
        """
        allowed = ("binary", "multi")
        if how not in allowed:
            raise ValueError(f"Illegal arg `how`. Got: {how}, Allowed: {allowed}")

        masks = self._read_files()

        res = []
        if how == "multi" and self.classes is not None:
            for c, i in list(self.classes.items())[1:]:
                args = []
                for true, pred, name in masks:
                    true_inst = true["inst_map"]
                    pred_inst = pred["inst_map"]
                    true_type = true["type_map"]
                    pred_type = pred["type_map"]

                    pred_type = get_type_instances(pred_inst, pred_type, i)
                    true_type = get_type_instances(true_inst, true_type, i)
                    args.append((true_type, pred_type, name, metrics, c))
                met = self.run_metrics(args, "inst", f"{c} instance seg")
                res.extend([metric for metric in met if metric])
        else:
            args = []
            for true, pred, name in masks:
                true = true["inst_map"]
                pred = pred["inst_map"]
                args.append((true, pred, name, metrics))
            met = self.run_metrics(args, "inst", "binary instance seg")
            res.extend([metric for metric in met if metric])

        return res

    def run_sem_benchmark(self, metrics: Tuple[str, ...] = ("iou",)) -> Dict[str, Any]:
        """Run semantic segmentation benchmarking.

        Parameters
        ----------
            metrics : Tuple[str, ...], default=("iou", )
                A tuple of the metrics that will be computed. Allowed metrics:
                "iou", "acc", "f1", "dice" "sensitivity", "specificity".

        Returns
        -------
            Dict[str, Any]:
                Dictionary mapping the metrics to values + metadata.
        """
        masks = self._read_files()

        args = []
        for true, pred, name in masks:
            true = true["sem_map"]
            pred = pred["sem_map"]
            args.append((true, pred, name, len(self.classes), metrics))
        met = self.run_metrics(args, "sem", "semantic seg")
        res = [metric for metric in met if metric]

        return res



from typing import Callable, Dict, List

import numpy as np
from torch.utils.data import Dataset

try:
    from ..transforms.albu_transforms import (
        IMG_TRANSFORMS,
        INST_TRANSFORMS,
        NORM_TRANSFORMS,
        apply_each,
        compose,
        to_tensorv3,
    )
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "To use the `csmp.dataset` module, the albumentations lib is needed. "
        "Install with `pip install albumentations`"
    )

__all__ = ["TrainDatasetBase"]


class TrainDatasetBase(Dataset):
    def __init__(
        self,
        img_transforms: List[str],
        inst_transforms: List[str],
        normalization: str = None,
        return_inst: bool = True,
        return_type: bool = True,
        return_sem: bool = False,
        return_weight: bool = False,
        **kwargs,
    ) -> None:
        """Training dataset baseclass.

        Parameters
        ----------
            img_transforms : List[str]
                A list containing all the transformations that are applied to the input
                images and corresponding masks. Allowed ones: "blur", "non_spatial",
                "non_rigid", "rigid", "hue_sat", "random_crop", "center_crop", "resize"
            inst_transforms : List[str]
                A list containg all the transformations that are applied to only the
                instance labelled masks. Allowed ones: "cellpose", "contour", "dist",
                "edgeweight", "hovernet", "omnipose", "smooth_dist", "binarize"
            normalization : str, optional
                Apply img normalization after all the transformations. One of "minmax",
                "norm", "percentile", None.
            return_inst : bool, default=True
                If True, returns a binarized instance mask. (If the db contains these.)
            return_type : bool, default=True
                If True, returns a type mask. (If the db contains these.)
            return_sem : bool, default=False
                If True, returns a semantic mask, (If the db contains these.)
            return_weight : bool, default=False
                Include a nuclear border weight map in the output.
            **kwargs:
                Arbitrary key-word arguments for the transformations.
        """
        super().__init__()

        allowed = list(IMG_TRANSFORMS.keys())
        if not all([tr in allowed for tr in img_transforms]):
            raise ValueError(
                f"Wrong img transformation. Got: {img_transforms}. Allowed: {allowed}."
            )
        allowed = list(NORM_TRANSFORMS.keys()) + [None]
        if normalization not in allowed:
            raise ValueError(
                f"Wrong norm transformation. Got: {normalization}. Allowed: {allowed}."
            )

        allowed = list(INST_TRANSFORMS.keys())
        if not all([tr in allowed for tr in inst_transforms]):
            raise ValueError(
                f"Wrong inst transformation. Got: {inst_transforms}. Allowed: {allowed}"
            )

        # Return masks
        self.return_inst = return_inst
        self.return_type = return_type
        self.return_sem = return_sem

        # Set transformations
        img_transforms = [IMG_TRANSFORMS[tr](**kwargs) for tr in img_transforms]
        if normalization is not None:
            img_transforms.append(NORM_TRANSFORMS[normalization]())

        inst_transforms = [INST_TRANSFORMS[tr](**kwargs) for tr in inst_transforms]
        if return_inst:
            inst_transforms.append(INST_TRANSFORMS["binarize"]())

        if return_weight:
            inst_transforms.append(INST_TRANSFORMS["edgeweight"]())

        self.img_transforms = compose(img_transforms)
        self.inst_transforms = apply_each(inst_transforms)
        self.to_tensor = to_tensorv3()

    def _getitem(
        self, ix: int, read_input_func: Callable, **kwargs
    ) -> Dict[str, np.ndarray]:
        """Get item.

        Parameters
        ----------
            ix : int
                An index for the iterable dataset.
            read_input_func : Callable
                A function that reads the images and corresponding masks and returns
                the inputs in a dictionary format. Return example:
                {"image": np.ndarray, "inst_map": np.ndarray, "type_map": np.ndarray}

        Returns
        -------
            Dict[str, np.ndarray]:
                A dictionary containing all the augmented data patches.
                Keys are: "im", "inst", "type", "sem". Image shape: (B, 3, H, W).
                Mask shapes: (B, C_mask, H, W).

        """
        inputs = read_input_func(ix, self.return_type, self.return_sem)

        # wrangle inputs to albumentations format
        mask_names = [key for key in inputs.keys() if key != "image"]
        masks = [arr for key, arr in inputs.items() if key != "image"]
        data = dict(image=inputs["image"], masks=masks)

        # transform + convert to tensor
        aug = self.img_transforms(**data)
        aux = self.inst_transforms(image=aug["image"], inst_map=aug["masks"][0])
        data = self.to_tensor(image=aug["image"], masks=aug["masks"], aux=aux)

        # wrangle data to return format
        out = dict(image=data["image"])
        for m, n in zip(data["masks"], mask_names):
            out[n] = m

        for n, aux_map in aux.items():
            out[n] = aux_map

        # remove redundant target (not needed in downstream).
        if self.return_inst:
            out["inst"] = out["binary"]
            del out["binary"]
        else:
            del out["inst"]

        return out



from pathlib import Path
from typing import Union

import numpy as np
import scipy.io

from .mask_utils import bounding_box, get_inst_centroid, get_inst_types

__all__ = ["mask2mat"]


def mask2mat(
    fname: Union[str, Path],
    save_dir: Union[str, Path],
    inst: np.ndarray,
    type: np.ndarray = None,
    sem: np.ndarray = None,
) -> None:
    """Convert one set of segmentation model output masks into a .mat file.

    Keys: "inst_map", "inst_type", "inst_centroid", "inst_bbox"
    Optional: "type_map", "sem_map",

    Parameters
    ----------
        fname : str or Path
            File name for the .mat file.
        save_dir : str or Path
            directory where the .mat files are saved.
        inst : np.ndarray
            Instance labelled instance segmentation mask from the segmentation model.
        sem : np.ndarray, optional
            Semantic segmentation mask from the segmentation model.
        type : np.ndarray, optional
            Cell type labelled semantic segmentation mask from the segmentation model.
    """
    save_dir = Path(save_dir)
    fname = Path(fname).with_suffix(".mat").name
    fn_mask = Path(save_dir / fname)

    if not Path(save_dir).exists():
        raise ValueError(f"`save_dir`: {save_dir} does not exist.")

    centroids = get_inst_centroid(inst)
    inst_ids = list(np.unique(inst)[1:])
    bboxes = np.array(
        [bounding_box(np.array(inst == id_, np.uint8)) for id_ in inst_ids]
    )

    mdict = {"inst_map": inst, "inst_centroids": centroids, "inst_bbox": bboxes}

    if type is not None:
        inst_types = get_inst_types(inst, type)
        mdict["inst_type"] = inst_types
        mdict["type_map"] = type

    if sem is not None:
        mdict["sem_map"] = sem

    scipy.io.savemat(file_name=fn_mask.as_posix(), mdict=mdict)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cellseg_models_pytorch',
 'cellseg_models_pytorch.datamodules',
 'cellseg_models_pytorch.datamodules.tests',
 'cellseg_models_pytorch.datasets',
 'cellseg_models_pytorch.datasets.dataset_writers',
 'cellseg_models_pytorch.datasets.tests',
 'cellseg_models_pytorch.decoders',
 'cellseg_models_pytorch.decoders.long_skips',
 'cellseg_models_pytorch.decoders.long_skips.merging',
 'cellseg_models_pytorch.decoders.tests',
 'cellseg_models_pytorch.inference',
 'cellseg_models_pytorch.inference.tests',
 'cellseg_models_pytorch.losses',
 'cellseg_models_pytorch.losses.criterions',
 'cellseg_models_pytorch.losses.tests',
 'cellseg_models_pytorch.metrics',
 'cellseg_models_pytorch.metrics.tests',
 'cellseg_models_pytorch.models',
 'cellseg_models_pytorch.models.base',
 'cellseg_models_pytorch.models.cellpose',
 'cellseg_models_pytorch.models.hovernet',
 'cellseg_models_pytorch.models.stardist',
 'cellseg_models_pytorch.models.tests',
 'cellseg_models_pytorch.modules',
 'cellseg_models_pytorch.modules.act',
 'cellseg_models_pytorch.modules.conv',
 'cellseg_models_pytorch.modules.norm',
 'cellseg_models_pytorch.modules.tests',
 'cellseg_models_pytorch.modules.upsample',
 'cellseg_models_pytorch.optimizers',
 'cellseg_models_pytorch.optimizers.tests',
 'cellseg_models_pytorch.postproc',
 'cellseg_models_pytorch.postproc.functional',
 'cellseg_models_pytorch.postproc.functional.cellpose',
 'cellseg_models_pytorch.postproc.tests',
 'cellseg_models_pytorch.training',
 'cellseg_models_pytorch.training.callbacks',
 'cellseg_models_pytorch.training.functional',
 'cellseg_models_pytorch.training.lit',
 'cellseg_models_pytorch.training.tests',
 'cellseg_models_pytorch.transforms',
 'cellseg_models_pytorch.transforms.albu_transforms',
 'cellseg_models_pytorch.transforms.functional',
 'cellseg_models_pytorch.transforms.tests',
 'cellseg_models_pytorch.utils',
 'cellseg_models_pytorch.utils.tests']

package_data = \
{'': ['*'],
 'cellseg_models_pytorch.datasets.tests': ['data/*',
                                           'data/imgs/*',
                                           'data/masks/*'],
 'cellseg_models_pytorch.inference.tests': ['data/*'],
 'cellseg_models_pytorch.training.tests': ['data/*'],
 'cellseg_models_pytorch.utils.tests': ['data/*']}

install_requires = \
['numba>=0.55.2,<0.56.0',
 'numpy>=1.21,<2.0',
 'opencv-python>=4.2.0.32,<5.0.0.0',
 'pathos>=0.2.8,<0.3.0',
 'scikit-image>=0.19,<0.20',
 'scikit-learn>=1.0.2,<2.0.0',
 'scipy>=1.7,<2.0',
 'timm>=0.5.4,<0.7.0',
 'torch>=1.8,<2.0',
 'torchvision>=0.9,<0.10',
 'tqdm>=4.64.0,<5.0.0']

extras_require = \
{'all': ['pytorch-lightning>=1.6.0,<2.0.0',
         'torch-optimizer>=0.3.0,<0.4.0',
         'tables>=3.6.0,<4.0.0',
         'albumentations>=1.0.0,<2.0.0',
         'requests>=2.28.0,<3.0.0']}

setup_kwargs = {
    'name': 'cellseg-models-pytorch',
    'version': '0.1.4',
    'description': 'Python library for 2D cell/nuclei instance segmentation models written with PyTorch.',
    'long_description': '<div align="center">\n\n![Logo](./images/logo.png)\n\n**Python library for 2D cell/nuclei instance segmentation models written with [PyTorch](https://pytorch.org/).**\n\n[![Generic badge](https://img.shields.io/badge/License-MIT-<COLOR>.svg?style=for-the-badge)](https://github.com/okunator/cellseg_models.pytorch/blob/master/LICENSE)\n[![PyTorch - Version](https://img.shields.io/badge/PYTORCH-1.8+-red?style=for-the-badge&logo=pytorch)](https://pytorch.org/)\n[![Python - Version](https://img.shields.io/badge/PYTHON-3.7+-red?style=for-the-badge&logo=python&logoColor=white)](https://www.python.org/)\n<br>\n[![Github Test](https://img.shields.io/github/workflow/status/okunator/cellseg_models.pytorch/Tests?label=Tests&logo=github&style=for-the-badge)](https://github.com/okunator/cellseg_models.pytorch/actions/workflows/tests.yml)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg?style=for-the-badge)](https://github.com/psf/black)\n[![Codecov](https://img.shields.io/codecov/c/github/okunator/cellseg_models.pytorch?logo=codecov&style=for-the-badge&token=oGSj7FZ1lm)](https://codecov.io/gh/okunator/cellseg_models.pytorch)\n<br>\n[![Pypi](https://img.shields.io/pypi/v/cellseg-models-pytorch?color=blue&logo=pypi&style=for-the-badge)](https://pypi.org/project/cellseg-models-pytorch/)\n<br>\n[![DOI](https://zenodo.org/badge/450787123.svg)](https://zenodo.org/badge/latestdoi/450787123)\n\n</div>\n\n<div align="center">\n\n</div>\n\n## Introduction\n\nContains multi-task encoder-decoder architectures along with dedicated post-processing methods for segmenting cell/nuclei instances. As the name suggests, this library is heavily inspired by [segmentation_models.pytorch](https://github.com/qubvel/segmentation_models.pytorch) library for semantic segmentation.\n\n<br><br>\n\n![Architecture](./images/architecture_overview.png)\n\n## Installation\n\n**Basic installation**\n\n```shell\npip install cellseg-models-pytorch\n```\n\n**To install extra dependencies (training utilities and datamodules for open-source datasets) use**\n\n```shell\npip install cellseg-models-pytorch[all]\n```\n\n## Features\n\n- High level API to define cell/nuclei instance segmentation models.\n- 4 cell/nuclei instance segmentation models and more to come.\n- Pre-trained backbones/encoders from the [timm](https://github.com/rwightman/pytorch-image-models) library.\n- All the architectures can be augmented to output semantic segmentation outputs along with instance semgentation outputs (panoptic segmentation).\n- A lot of flexibility to modify the components of the model architectures.\n- Optimized inference methods.\n- Popular training losses and benchmarking metrics.\n- Simple model training with [pytorch-lightning](https://www.pytorchlightning.ai/).\n- Popular optimizers for training (provided by [pytorch-optimizer](https://github.com/jettify/pytorch-optimizer)).\n\n## Models\n\n| Model                      | Paper                                                                          |\n| -------------------------- | ------------------------------------------------------------------------------ |\n| [[1](#Citation)] HoVer-Net | https://www.sciencedirect.com/science/article/pii/S1361841519301045?via%3Dihub |\n| [[2](#Citation)] Cellpose  | https://www.nature.com/articles/s41592-020-01018-x                             |\n| [[3](#Citation)] Omnipose  | https://www.biorxiv.org/content/10.1101/2021.11.03.467199v2                    |\n| [[4](#Citation)] Stardist  | https://arxiv.org/abs/1806.03535                                               |\n\n## Datasets\n\n| Dataset                       | Paper                                                                                            |\n| ----------------------------- | ------------------------------------------------------------------------------------------------ |\n| [[5, 6](#References)] Pannuke | https://arxiv.org/abs/2003.10778 , https://link.springer.com/chapter/10.1007/978-3-030-23937-4_2 |\n| [[7](#References)] Lizard     | http://arxiv.org/abs/2108.11195                                                                  |\n\n## Notebook examples\n\n- [Training Stardist with Pannuke](https://github.com/okunator/cellseg_models.pytorch/blob/main/examples/pannuke_nuclei_segmentation_stardist.ipynb). Train the Stardist model with constant sized Pannuke patches.\n- [Training Cellpose with Lizard](https://github.com/okunator/cellseg_models.pytorch/blob/main/examples/lizard_nuclei_segmentation_cellpose.ipynb). Train the Cellpose model with Lizard dataset that is composed of varying sized images.\n\n## Code Examples\n\n**Define Cellpose for cell segmentation.**\n\n```python\nimport cellseg_models_pytorch as csmp\nimport torch\n\nmodel = csmp.models.cellpose_base(type_classes=5) # num of cell types in training data=5.\nx = torch.rand([1, 3, 256, 256])\n\n# NOTE: these outputs still need post-processing to obtain instance segmentation masks.\ny = model(x) # {"cellpose": [1, 2, 256, 256], "type": [1, 5, 256, 256]}\n```\n\n**Define Cellpose for cell and tissue area segmentation (Panoptic segmentation).**\n\n```python\nimport cellseg_models_pytorch as csmp\nimport torch\n\nmodel = csmp.models.cellpose_plus(type_classes=5, sem_classes=3) # num cell types and tissue types\nx = torch.rand([1, 3, 256, 256])\n\n# NOTE: these outputs still need post-processing to obtain instance and semantic segmentation masks.\ny = model(x) # {"cellpose": [1, 2, 256, 256], "type": [1, 5, 256, 256], "sem": [1, 3, 256, 256]}\n```\n\n**Define panoptic Cellpose model with more flexibility.**\n\n```python\nimport cellseg_models_pytorch as csmp\n\nmodel = csmp.CellPoseUnet(\n    decoders=("cellpose", "sem"), # cellpose and semantic decoders\n    heads={"cellpose": {"cellpose": 2, "type": 5}, "sem": {"sem": 3}}, # three output heads\n    depth=5, # encoder depth\n    out_channels=(256, 128, 64, 32, 16), # number of out channels at each decoder stage\n    layer_depths=(4, 4, 4, 4, 4), # number of conv blocks at each decoder layer\n    style_channels=256, # Number of style vector channels\n    enc_name="resnet50", # timm encoder\n    enc_pretrain=True, # imagenet pretrained encoder\n    long_skip="unetpp", # use unet++ long skips. ("unet", "unetpp", "unet3p")\n    merge_policy="sum", # ("cat", "sum")\n    short_skip="residual", # residual short skips. ("basic", "residual", "dense")\n    normalization="bcn", # batch-channel-normalization. ("bcn", "bn", "gn", "ln", "in")\n    activation="gelu", # gelu activation instead of relu. Several options for this.\n    convolution="wsconv", # weight standardized conv. ("wsconv", "conv", "scaled_wsconv")\n    attention="se", # squeeze-and-excitation attention. ("se", "gc", "scse", "eca")\n    pre_activate=False, # normalize and activation after convolution.\n)\n\nx = torch.rand([1, 3, 256, 256])\n# NOTE: these outputs still need post-processing to obtain instance and semantic segmentation masks.\ny = model(x) # {"cellpose": [1, 2, 256, 256], "type": [1, 5, 256, 256], "sem": [1, 3, 256, 256]}\n```\n\n**Run HoVer-Net inference and post-processing with a sliding window approach.**\n\n```python\nimport cellseg_models_pytorch as csmp\n\nmodel = csmp.models.hovernet_base(type_classes=5)\n# returns {"hovernet": [B, 2, H, W], "type": [B, 5, H, W], "inst": [B, 2, H, W]}\n\n# Sliding window inference for big images using overlapping patches\ninferer = csmp.inference.SlidingWindowInferer(\n    model=model,\n    input_folder="/path/to/images/",\n    checkpoint_path="/path/to/model/weights/",\n    out_activations={"hovernet": "tanh", "type": "softmax", "inst": "softmax"},\n    out_boundary_weights={"hovernet": True, "type": False, "inst": False}, # smooths boundary effects\n    instance_postproc="hovernet", # THE POST-PROCESSING METHOD\n    patch_size=(256, 256),\n    stride=128,\n    padding=80,\n    batch_size=8,\n    normalization="percentile", # same normalization as in training\n)\n\ninferer.infer() # Run sliding window inference.\n\ninferer.out_masks\n# {"image1" :{"inst": [H, W], "type": [H, W]}, ..., "imageN" :{"inst": [H, W], "type": [H, W]}}\n```\n\n## Models API\n\n### Class API\n\nThe class API enables the most flexibility in defining different model architectures. It allows for defining a multitude of hard-parameter sharing multi-task encoder-decoder architectures with (relatively) low effort. The class API is borrowing a lot from [segmentation_models.pytorch](https://github.com/qubvel/segmentation_models.pytorch) models API.\n\n**Model classes**:\n\n- `csmp.CellPoseUnet`\n- `csmp.StarDistUnet`\n- `csmp.HoverNet`\n\n**All of the models contain**:\n\n- `model.encoder` - pretrained [timm](https://github.com/rwightman/pytorch-image-models) backbone for feature extraction.\n- `model.{decoder_name}_decoder` - Models can have multiple decoders with unique names.\n- `model.{head_name}_seg_head` - Model decoders can have multiple segmentation heads with unique names.\n- `model.forward(x)` - forward pass.\n\n**Defining your own multi-task architecture**\n\nFor example, to define a multi-task architecture that has `resnet50` encoder, four decoders, and 5 output heads with `CellPoseUnet` architectural components, we could do this:\n\n```python\nimport cellseg_models_pytorch as csmp\nimport torch\n\nmodel = csmp.CellPoseUnet(\n    decoders=("cellpose", "dist", "contour", "sem"),\n    heads={\n        "cellpose": {"type": 5, "cellpose": 2},\n        "dist": {"dist": 1},\n        "contour": {"contour": 1},\n        "sem": {"sem": 4}\n    },\n)\n\nx = torch.rand([1, 3, 256, 256])\nmodel(x)\n# {\n#   "cellpose": [1, 2, 256, 256],\n#   "type": [1, 5, 256, 256],\n#   "dist": [1, 1, 256, 256],\n#   "contour": [1, 1, 256, 256],\n#   "sem": [1, 4, 256, 256]\n# }\n```\n\n### Function API\n\nWith the function API, you can build models with low effort by calling the below listed functions. Under the hood, the function API simply calls the above classes with pre-defined decoder and head names. The training and post-processing tools of this library are built around these names, thus, it is recommended to use the function API, although, it is a bit more rigid than the class API. Basically, the function API only lacks the ability to define the output-tasks of the model, but allows for all the rest as the class API.\n\n| Model functions                        | Output names                              | Task                             |\n| -------------------------------------- | ----------------------------------------- | -------------------------------- |\n| `csmp.models.cellpose_base`            | `"type"`, `"cellpose"`,                   | **instance segmentation**        |\n| `csmp.models.cellpose_plus`            | `"type"`, `"cellpose"`, `"sem"`,          | **panoptic segmentation**        |\n| `csmp.models.omnipose_base`            | `"type"`, `"omnipose"`                    | **instance segmentation**        |\n| `csmp.models.omnipose_plus`            | `"type"`, `"omnipose"`, `"sem"`,          | **panoptic segmentation**        |\n| `csmp.models.hovernet_base`            | `"type"`, `"inst"`, `"hovernet"`          | **instance segmentation**        |\n| `csmp.models.hovernet_plus`            | `"type"`, `"inst"`, `"hovernet"`, `"sem"` | **panoptic segmentation**        |\n| `csmp.models.hovernet_small`           | `"type"`,`"hovernet"`                     | **instance segmentation**        |\n| `csmp.models.hovernet_small_plus`      | `"type"`, `"hovernet"`, `"sem"`           | **panoptic segmentation**        |\n| `csmp.models.stardist_base`            | `"stardist"`, `"dist"`                    | **binary instance segmentation** |\n| `csmp.models.stardist_base_multiclass` | `"stardist"`, `"dist"`, `"type"`          | **instance segmentation**        |\n| `csmp.models.stardist_plus`            | `"stardist"`, `"dist"`, `"type"`, `"sem"` | **panoptic segmentation**        |\n\n## References\n\n- [1] S. Graham, Q. D. Vu, S. E. A. Raza, A. Azam, Y-W. Tsang, J. T. Kwak and N. Rajpoot. "HoVer-Net: Simultaneous Segmentation and Classification of Nuclei in Multi-Tissue Histology Images." Medical Image Analysis, Sept. 2019.\n- [2] Stringer, C.; Wang, T.; Michaelos, M. & Pachitariu, M. Cellpose: a generalist algorithm for cellular segmentation Nature Methods, 2021, 18, 100-106\n- [3] Cutler, K. J., Stringer, C., Wiggins, P. A., & Mougous, J. D. (2022). Omnipose: a high-precision morphology-independent solution for bacterial cell segmentation. bioRxiv. doi:10.1101/2021.11.03.467199\n- [4] Uwe Schmidt, Martin Weigert, Coleman Broaddus, & Gene Myers (2018). Cell Detection with Star-Convex Polygons. In Medical Image Computing and Computer Assisted Intervention - MICCAI 2018 - 21st International Conference, Granada, Spain, September 16-20, 2018, Proceedings, Part II (pp. 265–273).\n- [5] Gamper, J., Koohbanani, N., Benet, K., Khuram, A., & Rajpoot, N. (2019) PanNuke: an open pan-cancer histology dataset for nuclei instance segmentation and classification. In European Congress on Digital Pathology (pp. 11-19).\n- [6] Gamper, J., Koohbanani, N., Graham, S., Jahanifar, M., Khurram, S., Azam, A.,Hewitt, K., & Rajpoot, N. (2020). PanNuke Dataset Extension, Insights and Baselines. arXiv preprint arXiv:2003.10778.\n- [7] Graham, S., Jahanifar, M., Azam, A., Nimir, M., Tsang, Y.W., Dodd, K., Hero, E., Sahota, H., Tank, A., Benes, K., & others (2021). Lizard: A Large-Scale Dataset for Colonic Nuclear Instance Segmentation and Classification. In Proceedings of the IEEE/CVF International Conference on Computer Vision (pp. 684-693).\n\n## Citation\n\n```bibtex\n@misc{csmp2022,\n    title={{cellseg_models.pytorch}: Cell/Nuclei Segmentation Models and Benchmark.},\n    author={Oskari Lehtonen},\n    howpublished = {\\url{https://github.com/okunator/cellseg_models.pytorch}},\n    doi = {10.5281/zenodo.7064617}\n    year={2022}\n}\n```\n\n## Licence\n\nThis project is distributed under [MIT License](https://github.com/okunator/cellseg_models.pytorch/blob/main/LICENSE)\n\nThe project contains code from the original cell segmentation and 3rd-party libraries that have permissive licenses:\n\n- [timm](https://github.com/rwightman/pytorch-image-models) (Apache-2)\n- [HoVer-Net](https://github.com/vqdang/hover_net) (MIT)\n- [Cellpose](https://github.com/MouseLand/cellpose) (BSD-3)\n- [Stardist](https://github.com/stardist/stardist) (BSD-3)\n\nIf you find this library useful in your project, it is your responsibility to ensure you comply with the conditions of any dependent licenses. Please create an issue if you think something is missing regarding the licenses.\n',
    'author': 'Okunator',
    'author_email': 'oskari.lehtonen@helsinki.fi',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/okunator/cellseg_models.pytorch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



from copy import deepcopy
from typing import Any, Dict, List

import torch
import torch.nn as nn
import yaml

try:
    import pytorch_lightning as pl
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "To use the `SegmentationExperiment`, pytorch-lightning is required. "
        "Install with `pip install pytorch-lightning`"
    )

from ...losses import JOINT_SEG_LOSSES, SEG_LOSS_LOOKUP, JointLoss, Loss, MultiTaskLoss
from ...optimizers import OPTIM_LOOKUP, SCHED_LOOKUP, adjust_optim_params
from ..callbacks import METRIC_LOOKUP


class SegmentationExperiment(pl.LightningModule):
    def __init__(
        self,
        model: nn.Module,
        branch_losses: Dict[str, int],
        branch_loss_params: Dict[str, Dict[str, Any]] = None,
        branch_metrics: Dict[str, List[str]] = None,
        optimizer: str = "adam",
        lookahead: bool = False,
        optim_params: Dict[str, Dict[str, Any]] = None,
        scheduler: str = "reduce_on_plateau",
        scheduler_params: Dict[str, Any] = None,
        log_freq: int = 100,
    ) -> None:
        """Segmentation model training experiment.

        Configures the training setup for a segmentation model.

        Parameters
        ----------
            model : nn.Module
                A segmentation model specification.
            branch_losses : Dict[str, str]
                A Dict of branch names mapped to a string specifying a jointloss.
                E.g. {"inst": "tversky_ce_ssim", "sem": "mse"}. Allowed losses: "mse",
                "ce", "sce", "ssim", "msssim", "tversky", "focal", "iou", "dice".
                Use underscores to create joint loss functions. e.g. "dice_ce_tversky".
            branch_loss_params : Dict[str, Dict[str, Any]], optional
                Params for the different losses at different branches. For example to
                use LS or class weighting when computing the losses.
                E.g. {"inst": {"apply_ls": True}, "sem": {"edge_weight": False}}
            branch_metrics : Dict[str, List[str]], optional
                A Dict of branch names mapped to a list of strings specifying a metrics.
                E.g. {"inst": ["acc"], "sem": ["miou", "acc"]}. Allowed metrics: "miou",
                "acc", "ssim", "iqi", "mse", None.
            optimizer : str, default="adam"
                Name of the optimizer. In-built optimizers from `torch` and
                `torch_optimizer` packages can be used. One of: "adam", "rmsprop","sgd",
                "adadelta", "apollo", "adabelief", "adamp", "adagrad", "adamax",
                "adamw", "asdg", "accsgd", "adabound", "adamod", "diffgrad", "lamb",
                "novograd", "pid", "qhadam", "qhm", "radam", "sgdw", "yogi", "ranger",
                "rangerqh","rangerva"
            optim_params : Dict[str, Dict[str, Any]]
                optim paramas like learning rates, weight decays etc for diff parts of
                the network.
                E.g. {"encoder": {"weight_decay: 0.1, "lr": 0.1}, "sem": {"lr": 0.01}}
            lookahead : bool, default=False
                Flag whether the optimizer uses lookahead.
            scheduler : str, default="reduce_on_plateau"
                The name of the learning rate scheduler (torch in-built schedulers).
                Allowed ones: "reduce_on_plateau", "lambda", "cyclic", "exponential",
                "cosine_annealing", "cosine_annealing_warm".
            scheduler_params : Dict[str, Any]
                Params dict for the scheduler. Refer to torch lr_scheduler docs
                for the possible scheduler arguments.
            return_soft_masks : bool, default=True
                Return the model outputs for logging if True. Saves mem if set to False.
            log_freq : int, default=100
                Return soft masks every every n batches for callbacks and logging.

        Raises
        ------
            ValueError if decoder branch related dicts don't have matching keys.
            ValueError if `branch_losses` contains illegal joint loss names.
            ValueError if illegal metric names are given.
            ValueError if illegal optimizer name is given.
            ValueError if illegal scheduler name is given.
        """
        super().__init__()
        self.model = model
        self.heads = model.heads
        self.aux_key = model.aux_key
        self.inst_key = model.inst_key

        self.optimizer = optimizer
        self.optim_params = optim_params
        self.scheduler = scheduler
        self.scheduler_params = scheduler_params
        self.lookahead = lookahead

        self.branch_losses = branch_losses
        self.branch_metrics = branch_metrics
        self.branch_loss_params = branch_loss_params
        self.log_freq = log_freq

        self._validate_branch_args()
        self.save_hyperparameters(ignore="model")

        self.criterion = self.configure_loss()
        metrics = self.configure_metrics()
        self.train_metrics = deepcopy(metrics)
        self.val_metrics = deepcopy(metrics)
        self.test_metrics = deepcopy(metrics)

    @classmethod
    def from_yaml(cls, model: nn.Module, yaml_path: str) -> pl.LightningModule:
        """Initialize the experiment from a yaml-file.

        Parameters
        ----------
            model : nn.Module
                Initialized model.
            yaml_path : str
                Path to the yaml file containing rest of the params
        """
        with open(yaml_path, "r") as stream:
            kwargs = yaml.full_load(stream)

        return cls(model, **kwargs)

    def forward(self, x: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Forward."""
        return self.model(x)

    def step(
        self, batch: Dict[str, torch.Tensor], batch_idx: int, phase: str
    ) -> Dict[str, torch.Tensor]:
        """General step."""
        soft_masks = self.model(batch["image"])
        targets = {k: val for k, val in batch.items() if k != "image"}

        loss = self.criterion(yhats=soft_masks, targets=targets)
        metrics = self.compute_metrics(soft_masks, targets, phase)

        ret = {
            "soft_masks": soft_masks if batch_idx % self.log_freq == 0 else None,
            "loss": loss,
        }

        return {**ret, **metrics}

    def training_step(
        self, batch: Dict[str, torch.Tensor], batch_idx: int
    ) -> Dict[str, torch.Tensor]:
        """Training step + train metric logs."""
        res = self.step(batch, batch_idx, "train")

        # log all the metrics
        self.log("train_loss", res["loss"], prog_bar=True, on_epoch=False, on_step=True)
        for k, val in res.items():
            if k not in ("loss", "soft_masks"):
                self.log(f"train_{k}", val, prog_bar=True, on_epoch=False, on_step=True)

        return res

    def validation_step(
        self, batch: Dict[str, torch.Tensor], batch_idx: int
    ) -> Dict[str, torch.Tensor]:
        """Validate step + validation metric logs + example outputs for logging."""
        res = self.step(batch, batch_idx, "val")

        # log all the metrics
        self.log("val_loss", res["loss"], prog_bar=False, on_epoch=True, on_step=False)
        for k, val in res.items():
            if k not in ("loss", "soft_masks"):
                self.log(f"val_{k}", val, prog_bar=False, on_epoch=True, on_step=False)

        return res

    def test_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> None:
        """Test step + test metric logs."""
        res = self.step(batch, batch_idx, "test")

        del res["soft_masks"]  # soft masks not needed for logging
        loss = res.pop("loss")

        # log all the metrics
        self.log("test_loss", loss, prog_bar=False, on_epoch=True, on_step=False)
        for k, val in res.items():
            self.log(f"test_{k}", val, prog_bar=False, on_epoch=True, on_step=False)

    def compute_metrics(
        self,
        preds: Dict[str, torch.Tensor],
        targets: Dict[str, torch.Tensor],
        phase: str,
    ) -> Dict[str, torch.Tensor]:
        """Compute metrics for logging."""
        if phase == "train":
            metrics_dict = self.train_metrics
        elif phase == "val":
            metrics_dict = self.val_metrics
        elif phase == "test":
            metrics_dict = self.test_metrics

        ret = {}
        for metric_name, metric in metrics_dict.items():
            if metric is not None:
                branch = metric_name.split("_")[0]
                ret[metric_name] = metric(preds[branch], targets[branch])

        return ret

    def _validate_branch_args(self) -> None:
        """Check that there are no conflicting decoder branch args."""
        lk = set(self.branch_losses.keys())
        dk = set(self.model._get_inner_keys(self.model.heads))
        has_same_keys = lk == dk

        mk = None
        if self.branch_metrics is not None:
            mk = set(self.branch_metrics.keys())
            has_same_keys = dk == lk == mk

        ek = None
        if self.branch_loss_params is not None:
            ek = set(self.branch_loss_params.keys())
            has_same_keys = dk == lk == mk == ek

        if not has_same_keys:
            raise ValueError(
                "Got mismatching keys for branch dict args. "
                f"Branch losses: {lk}. "
                f"Branch loss params: {ek}. "
                f"Decoder branches: {dk}. "
                f"Metrics: {mk}. "
                f"(`metrics`, and `branch_loss_params` can be None)"
            )

    def configure_loss(self) -> nn.Module:
        """Configure the single/multi-task loss for the model."""
        branch_losses = {}
        for branch, loss in self.branch_losses.items():
            if loss not in JOINT_SEG_LOSSES:
                raise ValueError(
                    f"Illegal joint loss given. Got: {loss}. Allowed are all the "
                    f"permutations of: {list(SEG_LOSS_LOOKUP.keys())}"
                )

            try:
                losses = loss.split("_")
            except Exception:
                losses = [loss]

            loss_kwargs = {"k": None}
            if self.branch_loss_params is not None:
                loss_kwargs = self.branch_loss_params[branch]

            # set joint losses for each branch
            branch_losses[branch] = JointLoss(
                [Loss(loss_name, **loss_kwargs) for loss_name in losses]
            )

        return MultiTaskLoss(branch_losses=branch_losses)

    def configure_metrics(self):
        """Configure the train, val, test metrics for logging."""
        allowed = list(METRIC_LOOKUP.keys()) + [None]
        metrics = nn.ModuleDict()
        for k, m in self.branch_metrics.items():
            for metric_name in m:
                if metric_name not in allowed:
                    raise ValueError(
                        f"Illegal metric given. Got: {metric_name}. Allowed: {allowed}."
                    )

                if metric_name is not None:
                    metric = METRIC_LOOKUP[metric_name]()
                else:
                    metric = None

                metrics[f"{k}_{metric_name}"] = metric

        return metrics

    def configure_optimizers(self):
        """Configure the optimizer and scheduler."""
        allowed = list(OPTIM_LOOKUP.keys())
        if self.optimizer not in allowed:
            raise ValueError(
                f"Illegal optimizer given. Got {self.optimizer}. Allowed: {allowed}."
            )

        allowed = list(SCHED_LOOKUP.keys())
        if self.scheduler not in allowed:
            raise ValueError(
                f"Illegal scheduler given. Got {self.scheduler}. Allowed: {allowed}."
            )

        # set sensible default if None.
        if self.optim_params is None:
            self.optim_params = {
                "encoder": {"lr": 0.00005, "weight_decay": 0.00003},
                "decoder": {"lr": 0.0005, "weight_decay": 0.0003},
            }

        params = adjust_optim_params(self.model, self.optim_params)
        optimizer = OPTIM_LOOKUP[self.optimizer](params)

        if self.lookahead:
            optimizer = OPTIM_LOOKUP["lookahead"](optimizer, k=5, alpha=0.5)

        if self.scheduler_params is None:
            self.scheduler_params = {}

        scheduler = {
            "scheduler": SCHED_LOOKUP[self.scheduler](
                optimizer, **self.scheduler_params
            ),
            "monitor": "val_loss",
            "interval": "epoch",
            "frequency": 1,
        }

        return [optimizer], [scheduler]



"""Imported most of the stuff from stardist repo. Minor modifications.

BSD 3-Clause License

Copyright (c) 2018-2022, Uwe Schmidt, Martin Weigert
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from typing import Tuple

import numpy as np
import scipy.ndimage as ndi
from skimage import img_as_ubyte
from skimage.draw import polygon
from skimage.measure import regionprops

from ...utils import bounding_box, remap_label, remove_small_objects
from .drfns import find_maxima, h_minima_reconstruction

__all__ = ["post_proc_stardist", "post_proc_stardist_orig", "polygons_to_label"]


def polygons_to_label_coord(
    coord: np.ndarray, shape: Tuple[int, int], labels: np.ndarray = None
) -> np.ndarray:
    """Render polygons to image given a shape.

    Parameters
    ----------
        coord.shape : np.ndarray
            Shape: (n_polys, n_rays)
        shape : Tuple[int, int]
            Shape of the output mask.
        labels : np.ndarray, optional
            Sorted indices of the centroids.

    Returns
    -------
        np.ndarray:
            Instance labelled mask. Shape: (H, W).
    """
    coord = np.asarray(coord)
    if labels is None:
        labels = np.arange(len(coord))

    assert coord.ndim == 3 and coord.shape[1] == 2 and len(coord) == len(labels)

    lbl = np.zeros(shape, np.int32)

    for i, c in zip(labels, coord):
        rr, cc = polygon(*c, shape)
        lbl[rr, cc] = i + 1

    return lbl


def ray_angles(n_rays: int = 32):
    """Get linearly spaced angles for rays."""
    return np.linspace(0, 2 * np.pi, n_rays, endpoint=False)


def dist_to_coord(
    dist: np.ndarray, points: np.ndarray, scale_dist: Tuple[int, int] = (1, 1)
) -> np.ndarray:
    """Convert list of distances and centroids from polar to cartesian coordinates.

    Parameters
    ----------
        dist : np.ndarray
            The centerpoint pixels of the radial distance map. Shape (n_polys, n_rays).
        points : np.ndarray
            The centroids of the instances. Shape: (n_polys, 2).
        scale_dist : Tuple[int, int], default=(1, 1)
            Scaling factor.

    Returns
    -------
        np.ndarray:
            Cartesian cooridnates of the polygons. Shape (n_polys, 2, n_rays).
    """
    dist = np.asarray(dist)
    points = np.asarray(points)
    assert (
        dist.ndim == 2
        and points.ndim == 2
        and len(dist) == len(points)
        and points.shape[1] == 2
        and len(scale_dist) == 2
    )
    n_rays = dist.shape[1]
    phis = ray_angles(n_rays)
    coord = (dist[:, np.newaxis] * np.array([np.sin(phis), np.cos(phis)])).astype(
        np.float32
    )
    coord *= np.asarray(scale_dist).reshape(1, 2, 1)
    coord += points[..., np.newaxis]
    return coord


def _ind_prob_thresh(prob: np.ndarray, prob_thresh: float, b: int = 2) -> np.ndarray:
    """Index based thresholding."""
    if b is not None and np.isscalar(b):
        b = ((b, b),) * prob.ndim

    ind_thresh = prob > prob_thresh
    if b is not None:
        _ind_thresh = np.zeros_like(ind_thresh)
        ss = tuple(
            slice(_bs[0] if _bs[0] > 0 else None, -_bs[1] if _bs[1] > 0 else None)
            for _bs in b
        )
        _ind_thresh[ss] = True
        ind_thresh &= _ind_thresh
    return ind_thresh.astype("int32")


def polygons_to_label(
    dist: np.ndarray,
    points: np.ndarray,
    shape: Tuple[int, int],
    prob: np.ndarray = None,
    thresh: float = -np.inf,
    scale_dist: Tuple[int, int] = (1, 1),
) -> np.ndarray:
    """Convert distances and center points to instance labelled mask.

    Parameters
    ----------
        dist : np.ndarray
            The centerpoint pixels of the radial distance map. Shape (n_polys, n_rays).
        points : np.ndarray
            The centroids of the instances. Shape: (n_polys, 2).
        shape : Tuple[int, int]:
            Shape of the output mask.
        prob : np.ndarray, optional
            The centerpoint pixels of the regressed distance transform.
            Shape: (n_polys, n_rays).
        thresh : float, default=-np.inf
            Threshold for the regressed distance transform.
        scale_dist : Tuple[int, int], default=(1, 1)
            Scaling factor.

    Returns
    -------
        np.ndarray:
            Instance labelled mask. Shape (H, W).
    """
    dist = np.asarray(dist)
    points = np.asarray(points)
    prob = np.inf * np.ones(len(points)) if prob is None else np.asarray(prob)

    assert dist.ndim == 2 and points.ndim == 2 and len(dist) == len(points)
    assert len(points) == len(prob) and points.shape[1] == 2 and prob.ndim == 1

    ind = prob > thresh
    points = points[ind]
    dist = dist[ind]
    prob = prob[ind]

    ind = np.argsort(prob, kind="stable")
    points = points[ind]
    dist = dist[ind]

    coord = dist_to_coord(dist, points, scale_dist=scale_dist)

    return polygons_to_label_coord(coord, shape=shape, labels=ind)


def _clean_up(inst_map: np.ndarray, size: int = 150, **kwargs) -> np.ndarray:
    """Clean up overlapping instances."""
    mask = remap_label(inst_map.copy())
    mask_connected = ndi.label(mask)[0]

    labels_connected = np.unique(mask_connected)[1:]
    for lab in labels_connected:
        inst = np.array(mask_connected == lab, copy=True)
        y1, y2, x1, x2 = bounding_box(inst)
        y1 = y1 - 2 if y1 - 2 >= 0 else y1
        x1 = x1 - 2 if x1 - 2 >= 0 else x1
        x2 = x2 + 2 if x2 + 2 <= mask_connected.shape[1] - 1 else x2
        y2 = y2 + 2 if y2 + 2 <= mask_connected.shape[0] - 1 else y2

        box_insts = mask[y1:y2, x1:x2]
        if len(np.unique(ndi.label(box_insts)[0])) <= 2:
            real_labels, counts = np.unique(box_insts, return_counts=True)
            real_labels = real_labels[1:]
            counts = counts[1:]
            max_pixels = np.max(counts)
            max_label = real_labels[np.argmax(counts)]
            for real_lab, count in list(zip(list(real_labels), list(counts))):
                if count < max_pixels:
                    if count < size:
                        mask[mask == real_lab] = max_label

    return mask


def post_proc_stardist(
    dist_map: np.ndarray, stardist_map: np.ndarray, thresh: float = 0.4, **kwargs
) -> np.ndarray:
    """Run post-processing for stardist.

    NOTE: This is not the original version that uses NMS.
    This is rather a workaround that is a little slower.

    Parameters
    ----------
        dist_map : np.ndarray
            Predicted distance transform. Shape: (H, W).
        stardist_map : np.ndarray
            Predicted radial distances. Shape: (n_rays, H, W).
        thresh : float, default=0.4
            Threshold for the regressed distance transform.

    Returns
    -------
        np.ndarray:
            Instance labelled mask. Shape: (H, W).
    """
    stardist_map = stardist_map.transpose(1, 2, 0)
    mask = _ind_prob_thresh(dist_map, thresh, b=2)

    # invert distmap
    inv_dist_map = 255 - img_as_ubyte(dist_map)

    # find markers from minima erosion reconstructed maxima of inv dist map
    reconstructed = h_minima_reconstruction(inv_dist_map)
    markers = find_maxima(reconstructed, mask=mask)
    markers = ndi.label(markers)[0]
    markers = remove_small_objects(markers, min_size=5)
    points = np.array(
        tuple(np.array(r.centroid).astype(int) for r in regionprops(markers))
    )

    if len(points) == 0:
        return np.zeros_like(mask)

    dist = stardist_map[tuple(points.T)]
    scores = dist_map[tuple(points.T)]

    labels = polygons_to_label(
        dist, points, prob=scores, shape=mask.shape, scale_dist=(1, 1)
    )
    labels = _clean_up(labels, **kwargs)

    return labels


def post_proc_stardist_orig(
    dist_map: np.ndarray, stardist_map: np.ndarray, thresh: float = 0.4, **kwargs
) -> np.ndarray:
    """Run the original stardist post-processing pipeline.

    NOTE: to use this the `stardist` package needs to be installed.

    Parameters
    ----------
        dist_map : np.ndarray
            Predicted distance transform. Shape: (H, W).
        stardist_map : np.ndarray
            Predicted radial distances. Shape: (n_rays, H, W).
        thresh : float, default=0.4
            Threshold for the regressed distance transform.

    Returns
    -------
        np.ndarray:
            Instance labelled mask. Shape: (H, W).
    """
    try:
        from stardist import non_maximum_suppression
    except ImportError:
        raise ImportError(
            "Install stardist if using the original post-proc method. "
            "`pip install stardist`"
        )
    rescale = (1, 1)
    img_shape = dist_map.shape
    stardist_map = stardist_map.transpose(1, 2, 0)
    points, probi, disti = non_maximum_suppression(
        stardist_map, dist_map, prob_thresh=thresh
    )
    labels = polygons_to_label(
        disti, points, prob=probi, shape=img_shape, scale_dist=rescale
    )

    return labels



import numpy as np
import pytest

from cellseg_models_pytorch.utils import (
    float2ubyte,
    minmax_normalize,
    normalize,
    percentile_normalize,
    percentile_normalize99,
)


@pytest.mark.parametrize("lower", [0.0, 0.1])
@pytest.mark.parametrize("upper", [99.99, 100.0])
def test_percentile_normalize(img_sample, lower, upper) -> None:
    im = img_sample
    nim = percentile_normalize(im, lower, upper)

    assert nim.dtype == "float32"
    assert nim.shape == im.shape
    assert np.logical_and(nim >= 0.0, nim <= 1.0).all()


@pytest.mark.parametrize("amin", [None, 0.0])
@pytest.mark.parametrize("amax", [1.0, None])
def test_percentile_normalize99(img_sample, amin, amax) -> None:
    im = img_sample
    nim = percentile_normalize99(im, amin, amax)

    assert nim.dtype == "float32"
    assert nim.shape == im.shape


@pytest.mark.parametrize("amin", [None, 0.0])
@pytest.mark.parametrize("amax", [1.0, None])
@pytest.mark.parametrize("standardize", [True, False])
def test_normalize(img_sample, standardize, amin, amax) -> None:
    im = img_sample
    nim = normalize(im, standardize, amin, amax)

    assert nim.dtype == "float32"
    assert nim.shape == im.shape


@pytest.mark.parametrize("amin", [None, 0.0])
@pytest.mark.parametrize("amax", [1.0, None])
def test_minmax_normalize(img_sample, amin, amax) -> None:
    im = img_sample
    nim = minmax_normalize(im, amin, amax)

    assert nim.dtype == "float32"
    assert nim.shape == im.shape
    assert np.logical_and(nim >= 0.0, nim <= 1.0).all()


@pytest.mark.parametrize("norm", [True, False])
def test_float2ubyte(img_sample, norm) -> None:
    im = img_sample
    nim = normalize(im)
    ubyte = float2ubyte(nim, norm)

    assert ubyte.dtype == "uint8"
    assert ubyte.shape == im.shape
    assert np.logical_and(ubyte >= 0, ubyte <= 255).all()



import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ["WSConv2d"]


# Adapted from
# https://github.com/joe-siyuan-qiao/WeightStandardization
class WSConv2d(nn.Conv2d):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        stride: int = 1,
        padding: int = 1,
        dilation: int = 1,
        groups: int = 1,
        bias: bool = True,
        eps: float = 1e-7,
        **kwargs
    ) -> None:
        """Weight standardized convolution.

        https://arxiv.org/abs/1903.10520

        Parameters
        ----------
            Refer to nn.Conv2d
        """
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
        )
        self.eps = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Weight standardized convolution forward pass."""
        weight = self.weight

        weight_mean = (
            weight.mean(dim=1, keepdim=True)
            .mean(dim=2, keepdim=True)
            .mean(dim=3, keepdim=True)
        )

        weight = weight - weight_mean
        std = weight.view(weight.size(0), -1).std(dim=1).view(-1, 1, 1, 1) + self.eps

        weight = weight / std.expand_as(weight)

        return F.conv2d(
            x, weight, self.bias, self.stride, self.padding, self.dilation, self.groups
        )



import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ["ScaledWSConv2d"]


# Adapted from:
# https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/layers/std_conv.py
class ScaledWSConv2d(nn.Conv2d):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        stride: int = 1,
        padding: int = 1,
        dilation: int = 1,
        groups: int = 1,
        bias: bool = True,
        gamma: float = 1.0,
        gain_init: float = 1.0,
        eps: float = 1e-7,
    ) -> None:
        """Conv2d layer with Scaled Weight Standardization.

        https://arxiv.org/abs/2101.08692

        Parameters
        ----------
            Refer to nn.Conv2d.

            gamma : float
                Fixed constant to preserve the variance of residual-blocks.
            gain_init : float
                Init value for the gain tensor.

        """
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
        )

        self.gain = nn.Parameter(torch.full((self.out_channels, 1, 1, 1), gain_init))
        self.scale = gamma * self.weight[0].numel() ** -0.5  # gamma * 1 / sqrt(fan-in)
        self.eps = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of scaled ws conv."""
        weight = F.batch_norm(
            self.weight.reshape(1, self.out_channels, -1),
            None,
            None,
            weight=(self.gain * self.scale).view(-1),
            training=True,
            momentum=0.0,
            eps=self.eps,
        ).reshape_as(self.weight)

        return F.conv2d(
            x, weight, self.bias, self.stride, self.padding, self.dilation, self.groups
        )



import torch.nn as nn

__all__ = ["initialize_decoder", "initialize_head"]


def initialize_decoder(module: nn.Module, activation: str = "relu") -> None:
    """Initialize a decoder module."""
    for m in module.modules():
        if isinstance(m, nn.Conv2d):
            if activation == "relu":
                nn.init.kaiming_uniform_(m.weight, mode="fan_in", nonlinearity="relu")
            else:
                nn.init.kaiming_normal_(m.weight, mode="fan_in")

            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

        if isinstance(m, nn.Conv1d):
            if activation == "relu":
                nn.init.kaiming_uniform_(m.weight, mode="fan_in", nonlinearity="relu")
            else:
                nn.init.kaiming_normal_(m.weight, mode="fan_in")

            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

        elif isinstance(m, nn.BatchNorm2d):
            nn.init.constant_(m.weight, 1)
            nn.init.constant_(m.bias, 0)

        elif isinstance(m, nn.Linear):
            nn.init.xavier_uniform_(m.weight)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)


def initialize_head(module: nn.Module):
    """Initialize a segmentation head module."""
    for m in module.modules():
        if isinstance(m, (nn.Linear, nn.Conv2d)):
            nn.init.xavier_uniform_(m.weight)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)



import torch
import torch.nn as nn

from .conv_base import (
    BasicConv,
    BasicConvOld,
    BottleneckConv,
    DepthWiseSeparableConv,
    FusedMobileInvertedConv,
    HoverNetDenseConv,
    InvertedBottleneckConv,
)
from .misc_modules import ChannelPool, StyleBlock

__all__ = ["ConvBlock"]

CONVBLOCK_LOOKUP = {
    "basic": BasicConv,
    "bottleneck": BottleneckConv,
    "dws": DepthWiseSeparableConv,
    "mbconv": InvertedBottleneckConv,
    "fmbconv": FusedMobileInvertedConv,
    "hover_dense": HoverNetDenseConv,
    "basic_old": BasicConvOld,
}


class ShortSkipMixIn:
    """Add short skip functionality.

    - Dense short skip forward method
    - Residual short skip forward method
    - Forward method w/o short skip.
    """

    def forward_basic(self, x: torch.Tensor) -> torch.Tensor:
        """Apply a generic forward pass without short skip connection."""
        if self.block.preactivate:
            out = self.block.forward_features_preact(x)
        else:
            out = self.block.forward_features(x)

        return out

    def forward_residual(self, x: torch.Tensor) -> torch.Tensor:
        """Apply a generic forward pass with residual short skip."""
        identity = x
        if self.downsample is not None:
            identity = self.downsample(x)

        if self.block.preactivate:
            out = self.block.forward_features_preact(x)
        else:
            out = self.block.forward_features(x)

        out = out + identity

        return out

    def forward_dense(self, x: torch.Tensor) -> torch.Tensor:
        """Apply a generic forward pass with dense short skip connection."""
        prev_features = [x] if isinstance(x, torch.Tensor) else x
        x = torch.cat(prev_features, dim=1)

        if self.block.preactivate:
            out = self.block.forward_features_preact(x)
        else:
            out = self.block.forward_features(x)

        return out


class ConvBlock(nn.Module, ShortSkipMixIn):
    def __init__(
        self,
        name: str,
        in_channels: int,
        out_channels: int,
        style_channels: int = None,
        short_skip: str = "residual",
        same_padding: bool = True,
        normalization: str = "bn",
        activation: str = "relu",
        convolution: str = "conv",
        preactivate: bool = False,
        kernel_size: int = 3,
        groups: int = 1,
        bias: bool = True,
        attention: str = None,
        preattend: bool = False,
        use_style: bool = True,
        **kwargs,
    ) -> None:
        """Wrap different conv-blocks under one generic conv-block.

        Adds one channel pooling block if `short_skip` == "residual".

        Optional:
            - add a style vector to the output at the end of the block (Cellpose).

        Parameters
        ----------
            name : str
                The name of the conv-block. One of: "basic". "mbconv", "fmbconv" "dws",
                "bottleneck".
            in_channels : int
                Number of input channels.
            out_channels : int
                Number of output channels.
            style_channels : int, default=None
                Number of style vector channels. If None, style vectors are ignored.
            short_skip : str, default="residual"
                The name of the short skip method. One of: "residual", "dense", "basic"
            same_padding : bool, default=True
                if True, performs same-covolution.
            normalization : str, default="bn":
                Normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", "lrn", None
            activation : str, default="relu"
                Activation method.
                One of: "mish", "swish", "relu", "relu6", "rrelu", "selu",
                "celu", "gelu", "glu", "tanh", "sigmoid", "silu", "prelu",
                "leaky-relu", "elu", "hardshrink", "tanhshrink", "hardsigmoid"
            convolution : str, default="conv"
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            preactivate : bool, default=False
                If True, normalization will be applied before convolution.
            kernel_size : int, default=3
                The size of the convolution kernel.
            groups : int, default=1
                Number of groups the kernels are divided into. If `groups == 1`
                normal convolution is applied. If `groups = in_channels`
                depthwise convolution is applied.
            bias : bool, default=True,
                Include bias term in the convolution block. Only used for `BaasicConv`.
            attention : str, default=None
                Attention method. One of: "se", "scse", "gc", "eca", None
            preattend : bool, default=False
                If True, Attention is applied at the beginning of forward pass.
            use_style : bool, default=False
                If True and `style_channels` is not None, adds a style vec to output.

        Raises
        ------
            ValueError:
                - If illegal `name`is given as input argument.
                - If illegal `short_skip` is given as input argument.
        """
        super().__init__()
        self.short_skip = short_skip

        allowed = list(CONVBLOCK_LOOKUP.keys())
        if name not in allowed:
            raise ValueError(
                f"Illegal convblock name given. Got '{name}'. Allowed: {allowed}"
            )

        allowed = ("residual", "dense", "basic")
        if short_skip not in allowed:
            raise ValueError(
                f"Illegal `short_skip` given. Got: '{short_skip}'. Allowed: {allowed}."
            )

        self.block = CONVBLOCK_LOOKUP[name](
            in_channels=in_channels,
            out_channels=out_channels,
            same_padding=same_padding,
            normalization=normalization,
            activation=activation,
            convolution=convolution,
            preactivate=preactivate,
            kernel_size=kernel_size,
            groups=groups,
            bias=bias,
            attention=attention,
            preattend=preattend,
            **kwargs,
        )

        self.downsample = None
        if short_skip == "residual" and in_channels != self.out_channels:
            self.downsample = ChannelPool(
                in_channels=in_channels,
                out_channels=self.out_channels,
                convolution=convolution,
                normalization=normalization,
            )

        self.add_style = None
        if style_channels is not None and use_style:
            self.add_style = StyleBlock(style_channels, out_channels)

    @property
    def out_channels(self) -> int:
        """Set out_channels."""
        return self.block.out_channels

    def forward(self, x: torch.Tensor, style: torch.Tensor = None) -> torch.Tensor:
        """Forward pass of the conv-block."""
        if self.short_skip == "residual":
            x = self.forward_residual(x)
        elif self.short_skip == "dense":
            x = self.forward_dense(x)
        else:
            x = self.forward_basic(x)

        if self.add_style is not None:
            x = self.add_style(x, style)

        return x
